

kernel mandelbrot : Spoc.Vector.vint32 -> int -> int -> int -> unit = "kernels/Mandelbrot" "mandelbrot"


let devices = Spoc.Devices.init ();;

let dev = ref devices.(0);;


let auto_transfers = ref true;;


let largeur = ref 1024;;
let hauteur = ref 1024;;

let max_iter  = ref 500;;


let b_iter = Spoc.Vector.create Spoc.Vector.int32(*`Int32 Bigarray.int32*) ((!largeur)*(!hauteur));;


let img = Array.make_matrix !largeur !hauteur Graphics.black;;



let get_min ar =
	Spoc.Mem.to_cpu ar ();
	let min = ref !max_iter in
	for i = 0 to (Spoc.Vector.length ar - 1) do
		if (Int32.to_int (Spoc.Mem.unsafe_get ar i)) > 1 then
		if (Int32.to_int (Spoc.Mem.unsafe_get ar i)) < !min then
			min :=  (Int32.to_int (Spoc.Mem.unsafe_get ar i))
	done;
	!min

let couleur n = 
	if n =  !max_iter then
		Graphics.black
	else let f n = 
		let i = float n in 
		int_of_float (255. *. (0.5 +. 0.5 *. sin(i *. 0.1))) in
	 Graphics.rgb (*(f (n + 32))*)0  (f(n + 16))  (f n)


let main  = 
	let arg1 = ("-device" , Arg.Int (fun i  -> dev := devices.(i)), "number of the device [0]")
	and arg2 = ("-width" , Arg.Int (fun i  -> largeur := i), "width of the image to compute [1024]")
	and arg3 = ("-height" , Arg.Int (fun i  -> hauteur := i), "height of the image to compute [1024]")
	and arg4 = ("-max_iter" , Arg.Int (fun b -> max_iter := b), "max number of iterations [1024]") in
	Arg.parse ([arg1;arg2;arg3;arg4]) (fun s -> ()) "";
	Printf.printf "Will use device : %s\n" (!dev).Spoc.Devices.general_info.Spoc.Devices.name;
	let threadsPerBlock = 256 in
	let blocksPerGrid = (!largeur* !hauteur + threadsPerBlock -1) / threadsPerBlock in
	let block = {Spoc.Kernel.blockX = threadsPerBlock; Spoc.Kernel.blockY = 1; Spoc.Kernel.blockZ = 1}
	and grid= {Spoc.Kernel.gridX = blocksPerGrid; Spoc.Kernel.gridY = 1; Spoc.Kernel.gridZ = 1} in
	Spoc.Kernel.run !dev (block,grid)(*exec ~cached:true*) (mandelbrot  b_iter !max_iter !largeur !hauteur);
  let l = string_of_int !largeur in
  let h = string_of_int !hauteur in
  let dim = " " ^l^"x"^h in

	Spoc.Mem.to_cpu b_iter ();
		Spoc.Devices.flush !dev ();

	Graphics.open_graph dim;
  for a = 0 to pred (!largeur - 1 ) do
    for b = 0 to pred (!hauteur - 1 )do
      img.(a).(b) <-  (couleur (Int32.to_int ((Spoc.Mem.unsafe_get b_iter (b * !hauteur + a) )))) ;
    done;
	done;
  Graphics.draw_image (Graphics.make_image img) 0 0;
 ignore (Graphics.read_key ())

	

