open Unix

kernel mandelbrot : Spoc.Vector.vint32 -> int -> int -> int -> int -> int -> unit = "kernels/Mandelbrot2" "mandelbrot"


let devices = Spoc.Devices.init ()

let dev1 = ref devices.(0);;
let dev2 = ref devices.(0);;


let auto_transfers = ref true;;


let largeur = ref 1024;;
let hauteur = ref 1024;;

let max_iter  = ref 500;;


let b_iter = Spoc.Vector.create Spoc.Vector.int32(*`Int32 Bigarray.int32*) ((!largeur)*(!hauteur));;


let img = Array.make_matrix !largeur !hauteur Graphics.black;;



let get_min ar =
	Spoc.Mem.to_cpu ar ();
	let min = ref !max_iter in
	for i = 0 to (Spoc.Vector.length ar - 1) do
		if (Int32.to_int (Spoc.Mem.unsafe_get ar i)) > 1 then
		if (Int32.to_int (Spoc.Mem.unsafe_get ar i)) < !min then
			min :=  (Int32.to_int (Spoc.Mem.unsafe_get ar i))
	done;
	!min

let couleur n = 
	if n =  !max_iter then
		Graphics.black
	else let f n = 
		let i = float n in 
		int_of_float (255. *. (0.5 +. 0.5 *. sin(i *. 0.1))) in
	 Graphics.rgb (*(f (n + 32))*)0  (f(n + 16))  (f n)


let main  = 
	let arg0 = ("-device1" , Arg.Int (fun i  -> dev1 := devices.(i)), "number of the device [0]")
	and arg1 = ("-device2" , Arg.Int (fun i  -> dev2 := devices.(i)), "number of the device [0]")
	and arg2 = ("-width" , Arg.Int (fun i  -> largeur := i), "width of the image to compute [1024]")
	and arg3 = ("-height" , Arg.Int (fun i  -> hauteur := i), "height of the image to compute [1024]")
	and arg4 = ("-max_iter" , Arg.Int (fun b -> max_iter := b), "max number of iterations [1024]") in
	Arg.parse ([arg0; arg1;arg2;arg3;arg4]) (fun s -> ()) "";
	Printf.printf "Will use device : %s and %s\n" (!dev1).Spoc.Devices.general_info.Spoc.Devices.name (!dev2).Spoc.Devices.general_info.Spoc.Devices.name;
  Pervasives.flush Pervasives.stdout;
	let threadsPerBlock = 256 in
	let blocksPerGrid = (!largeur* !hauteur + threadsPerBlock -1) / threadsPerBlock in
	let block = {Spoc.Kernel.blockX = threadsPerBlock; Spoc.Kernel.blockY = 1; Spoc.Kernel.blockZ = 1}
	and grid= {Spoc.Kernel.gridX = blocksPerGrid/2; Spoc.Kernel.gridY = 1; Spoc.Kernel.gridZ = 1} in
	let b_iter1 = Spoc.Mem.sub_vector b_iter 0 (!largeur* !hauteur/2) 
	and b_iter2 = Spoc.Mem.sub_vector b_iter (!largeur* !hauteur/2) (!largeur* !hauteur/2) in 
	
	Spoc.Kernel.run !dev1 (block,grid) (*exec ~cached:true*) (mandelbrot  b_iter1 !max_iter !largeur (!hauteur) 0 (!hauteur/2));
	Spoc.Kernel.run !dev2 (block,grid) (*exec ~cached:true*) (mandelbrot  b_iter2 !max_iter !largeur (!hauteur) (!hauteur/2) (!hauteur));

	  let l = string_of_int !largeur in
  let h = string_of_int !hauteur in
  let dim = " " ^l^"x"^h in
	Spoc.Mem.to_cpu b_iter1 ();
	Spoc.Mem.to_cpu b_iter2 ();
	Spoc.Devices.flush !dev1 (); 
	Spoc.Devices.flush !dev2 (); 

  for a = 0 to pred (!largeur - 1 ) do
    for b = 0 to pred (!hauteur - 1 )do
      img.(a).(b) <-  (couleur (Int32.to_int ((Spoc.Mem.unsafe_get b_iter (a * !hauteur + b) )))) ;
    done;
	done;
  Graphics.open_graph dim;
  Graphics.draw_image (Graphics.make_image img) 0 0;
 ignore (Graphics.read_key ())

	

