open Spoc			

let dimA = (4096,4096)
let dimB = (4096,4096)
let dimC = (4096,4096)

kernel ati_matmult : Spoc.Vector.vfloat32 -> Spoc.Vector.vfloat32 -> Spoc.Vector.vfloat32 -> int -> Spoc.Vector.localvfloat32 -> unit = "kernels/MatrixMultiplication_Kernels" "mmmKernel_local"


let sgemm_with_cublas devices =
	let dev = devices.(0) in
	Printf.printf "With %s\n" dev.Devices.general_info.Devices.name;
	let t0 = Unix.gettimeofday () in
		let a = Vector.create Vector.float32 (fst dimA * snd dimA) 
		and b = Vector.create Vector.float32 (fst dimB * snd dimB) 
		and c = Vector.create Vector.float32 (fst dimC * snd dimC) in

		for i = 0 to (Vector.length a - 1) do
				Mem.unsafe_set a i ((Random.float 255.));
				Mem.unsafe_set b i ((Random.float 255.));
		done;
		let t1 = Unix.gettimeofday () in
for i = 0 to 99 do	
		Cublas.run dev (Cublas.cublasSgemm 'N' 'N' (snd dimA)(snd dimC) (snd dimB)
				1. a (fst dimA) b (fst dimB) 1. c (fst dimC)) ;
		Mem.to_cpu c ();
		Mem.to_cpu a ();
		Mem.to_cpu b ();
		Devices.flush dev ();
done;
		let t2 = Unix.gettimeofday () in
			Printf.printf "total_time : %gs - compute time : %gs - %gGFlops\n" (t2-.t0) (t2-.t1) (100.*.(2.*.(float)((snd dimA)*(snd dimC)*(snd dimB))	)	/.(t2-.t1)/.1.e09)
			

let sgemm_with_ati_sdk devices =
	let dev = devices.(3) in
	Printf.printf "With %s\n" dev.Devices.general_info.Devices.name;
	Pervasives.flush Pervasives.stdout;
	
	let t0 = Unix.gettimeofday () in
	let a = Vector.create Vector.float32 (fst dimA * snd dimA) 
	and b = Vector.create Vector.float32 (fst dimB * snd dimB) 
	and c = Vector.create Vector.float32 (fst dimC * snd dimC) 
	and tmpa = Vector.create Vector.float32 (4*8*8) 
	and tmpb = Vector.create Vector.float32 10 
	in
	for i = 0 to (Vector.length a - 1) do
	  Mem.unsafe_set a i ((Random.float 2.));
	  Mem.unsafe_set b i ((Random.float 2.));
	done;
	let open Spoc.Kernel in
  	let block = {blockX = 8; blockY = 8; blockZ = 1}
  	and grid = {gridX = 4096/4/8;  gridY = 4096/4/8; gridZ = 1} in
			let t1 = Unix.gettimeofday () in
			for i = 0 to 99 do	
	      Kernel.run dev (block, grid)  ~cached:true (ati_matmult a b c (snd dimA) tmpa );
			done;
		  Mem.to_cpu c ();
      Devices.flush dev ();
	let t2 = Unix.gettimeofday () in
	Printf.printf "total_time : %gs - compute time : %gs - %gGFlops\n" (t2-.t0) (t2-.t1)
	  (100.*.(2.*.(float)((snd dimA)*(snd dimC)*(snd dimB)))/.(t2-.t1)/.1.e09)
	  

let multi_matmul devices = 
	let dimA = (fst dimA,snd dimA*2) in
	let dimB = (4096*2,4096) in
	let dimC = (4096,4096*2) in
	let ati_dev = devices.(3) 
	and cuda_dev = devices.(0) in
	Printf.printf "With %s and %s\n" ati_dev.Devices.general_info.Devices.name  cuda_dev.Devices.general_info.Devices.name;
	Pervasives.flush Pervasives.stdout;
	let t0 = Unix.gettimeofday () in
	let a = Vector.create Vector.float32 (fst dimA * snd dimA) 
	and b = Vector.create Vector.float32 (fst dimB * snd dimB) 
	and c = Vector.create Vector.float32 (fst dimC * snd dimC) 
	and c_verify = Vector.create Vector.float32 (fst dimC * snd dimC) 
	in
	for i = 0 to (Vector.length a - 1) do
	  Mem.unsafe_set a i ((Random.float 2.));
	done;
	for i = 0 to (Vector.length b - 1) do
	  Mem.unsafe_set b i ((Random.float 2.));	
	done;
	let a1 = Mem.sub_vector a 0 ((fst dimA * snd dimA)/2)
	and a2 = Mem.sub_vector a ((fst dimA * snd dimA)/2) ((fst dimA * snd dimA)/2)
	and c1 = Mem.sub_vector c 0 ((fst dimC * snd dimC)/2)
	and c2 = Mem.sub_vector c ((fst dimC * snd dimC)/2) ((fst dimC * snd dimC)/2)
	and b1 = Mem.sub_vector b 0 ((fst dimB * snd dimB))
	and b2 = Mem.sub_vector b 0 ((fst dimB * snd dimB))
	and tmpa1 = Vector.create Vector.float32 (4*8*8)
	and tmpb1 = Vector.create Vector.float32 10
	and t1 = Unix.gettimeofday () in
	for i = 0 to 99 do	
	  let open Spoc.Kernel in
	      let block = {blockX = 8; blockY = 8; blockZ = 1}
				and grid = {gridX = 4096/4/8;  gridY = 4096/4/8; gridZ = 1} in
		Kernel.run ati_dev (block, grid) ~cached:true (ati_matmult a1 b1 c1 (snd dimA/2) tmpa1 );
	(*    Kernel.run ati_dev (block, grid) (ati_matmult c1 a1 b1  ((snd dimA)) (fst dimB) );*)
	    Cublas.run cuda_dev (Cublas.cublasSgemm 'N' 'N' ((snd dimA)/2) ((snd dimC)/2) ((snd dimB))
				   1. a2 (fst dimA) b2 (fst dimB) 1. c2 (fst dimC)) ;

	    Mem.to_cpu c1 ();
	    Mem.to_cpu c2 ();
	    Mem.to_cpu a1 ();
	    Mem.to_cpu a2 ();
	    Mem.to_cpu b1 ();
	    Mem.to_cpu b2 ();
	    Devices.flush ati_dev ();
	    Devices.flush cuda_dev ();
	done;	    
	let t2 = Unix.gettimeofday () in
	Printf.printf "total_time : %gs - compute time : %gs - %gGFlops\n" (t2-.t0) (t2-.t1)
	  (100.*.(2.*.(float)((snd dimA)*(snd dimC)*(snd dimB)))/.(t2-.t1)/.1.e09)


	      

let _ =
	Random.self_init();

	let devices = Devices.init () in
	
	Cublas.init ();

	sgemm_with_cublas devices;
	sgemm_with_ati_sdk devices;
	multi_matmul devices;

	Cublas.shutdown ();

;;
