type kernel
external relax_vcustom : 'a Vector.vcustom -> 'b Vector.vcustom = "%identity"
type ('a, 'b) kernelArgs =
    VCustom of 'a Vector.vcustom
  | VFloat32 of Vector.vfloat32
  | VFloat64 of Vector.vfloat64
  | VComplex32 of Vector.vcomplex32
  | VInt32 of Vector.vint32
  | VInt64 of Vector.vint64
  | Int32 of int
  | Int64 of int
  | Float32 of float
  | Float64 of float
  | Vector of ('a, 'b) Vector.vector
  | LocalFloat32 of Vector.localvfloat32
type block = {
  mutable blockX : int;
  mutable blockY : int;
  mutable blockZ : int;
}
type grid = {
  mutable gridX : int;
  mutable gridY : int;
  mutable gridZ : int;
}
val nbCaches : int
module Cuda :
  sig
    type cuda_extra
    external cuda_create_extra : int -> cuda_extra = "spoc_cuda_create_extra"
    external cuda_load_param_vec :
      int ref -> cuda_extra -> Vector.device_vec -> unit
      = "spoc_cuda_load_param_vec_b" "spoc_cuda_load_param_vec_n"
    external cuda_custom_load_param_vec :
      int ref -> cuda_extra -> Vector.device_vec -> unit
      = "spoc_cuda_custom_load_param_vec_b"
      "spoc_cuda_custom_load_param_vec_n"
    external cuda_load_param_int : int ref -> cuda_extra -> int -> unit
      = "spoc_cuda_load_param_int_b" "spoc_cuda_load_param_int_n"
    external cuda_load_param_int64 : int ref -> cuda_extra -> int -> unit
      = "spoc_cuda_load_param_int64_b" "spoc_cuda_load_param_int64_n"
    external cuda_load_param_float : int ref -> cuda_extra -> float -> unit
      = "spoc_cuda_load_param_float_b" "spoc_cuda_load_param_float_n"
    external cuda_load_param_float64 : int ref -> cuda_extra -> float -> unit
      = "spoc_cuda_load_param_float64_b" "spoc_cuda_load_param_float64_n"
    external set_block_shape : kernel -> block -> Devices.generalInfo -> unit
      = "spoc_cuda_set_block_shape"
    external cuda_launch_grid :
      int ref ->
      kernel -> grid -> block -> cuda_extra -> Devices.generalInfo -> unit
      = "spoc_cuda_launch_grid_b" "spoc_cuda_launch_grid_n"
    external cuda_create_dummy_kernel : unit -> kernel
      = "spoc_cuda_create_dummy_kernel"
    val cudaKernelCache :
      (Digest.t, Devices.generalInfo -> kernel) Hashtbl.t array ref
    val cuda_load :
      bool ->
      int ->
      string ->
      string -> ('a, 'b) kernelArgs array -> Devices.generalInfo -> kernel
    val cuda_load_arg :
      int ref ->
      cuda_extra -> Devices.device -> 'a -> 'b -> ('c, 'd) kernelArgs -> unit
  end
module OpenCL :
  sig
    external opencl_load_param_vec :
      int ref ->
      kernel -> int -> Vector.device_vec -> Devices.generalInfo -> unit
      = "spoc_opencl_load_param_vec"
    external opencl_load_param_local_vec :
      int ref ->
      kernel -> int -> Vector.device_vec -> Devices.generalInfo -> unit
      = "spoc_opencl_load_param_local_vec"
    external opencl_load_param_int :
      int ref -> kernel -> int -> Devices.generalInfo -> unit
      = "spoc_opencl_load_param_int"
    external opencl_load_param_int64 :
      int ref -> kernel -> int -> Devices.generalInfo -> unit
      = "spoc_opencl_load_param_int64"
    external opencl_load_param_float :
      int ref -> kernel -> float -> Devices.generalInfo -> unit
      = "spoc_opencl_load_param_float"
    external opencl_load_param_float64 :
      int ref -> kernel -> float -> Devices.generalInfo -> unit
      = "spoc_opencl_load_param_float64"
    external opencl_launch_grid :
      kernel -> grid -> block -> Devices.generalInfo -> int -> unit
      = "spoc_opencl_launch_grid"
    external opencl_create_dummy_kernel : unit -> kernel
      = "spoc_opencl_create_dummy_kernel"
    val openCLKernelCache :
      (Digest.t, Devices.generalInfo -> kernel) Hashtbl.t array ref
    val opencl_load :
      bool -> int -> string -> string -> Devices.generalInfo -> kernel
    val opencl_load_arg :
      int ref ->
      Devices.device -> kernel -> int -> ('a, 'b) kernelArgs -> unit
  end
exception ERROR_BLOCK_SIZE
exception ERROR_GRID_SIZE
val exec :
  string ->
  string ->
  ('a, 'b) kernelArgs array ->
  block * grid -> bool -> int -> Devices.device -> unit
val run :
  Devices.device ->
  block * grid ->
  ?cached:bool ->
  ?queue_id:int ->
  (block * grid -> bool -> int -> Devices.device -> 'a) -> 'a
val max : 'a -> 'a -> 'a
