#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/custom.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <caml/bigarray.h>
#include <math.h>
#include <string.h>
#include "Spoc.h"
/**************** KERNEL ******************/

int ae_load_file_to_memory(const char *filename, char **result)
{
	int size = 0;
	FILE *f = fopen(filename, "rb");
	if (f == NULL)
	{
		*result = NULL;
		return -1; // -1 means file opening fail
	}
	fseek(f, 0, SEEK_END);
	size = ftell(f);
	fseek(f, 0, SEEK_SET);
	*result = (char *)malloc(size+1);
	if (size != fread(*result, sizeof(char), size, f))
	{
		free(*result);
		return -2; // -2 means file reading fail
	}
	fclose(f);
	(*result)[size] = 0;
	return size;
}


CAMLprim value spoc_cuda_load(value moduleSrc, value function_name, value args, value gi){
	CAMLparam4(moduleSrc, function_name, args, gi);
	CUmodule module;
	CUfunction *kernel;
	char* functionN;
	char *ptx_source;
	const unsigned int jitNumOptions = 3;

	CUjit_option jitOptions[3];
	void *jitOptVals[3];
	int jitLogBufferSize;
	char *jitLogBuffer;
	int jitRegCount = 32;

	CUDA_GET_CONTEXT;


	kernel = malloc(sizeof(CUfunction));
//	moduleN = String_val(module_name);
	functionN = String_val(function_name);
//	ptx_fname = malloc(sizeof(char)*(strlen(moduleN)+4));
//	strcpy (ptx_fname, moduleN);
//	strcat(ptx_fname, ".ptx");
//
//	err = ae_load_file_to_memory(ptx_fname, &ptx_source);
	ptx_source = String_val(moduleSrc);
//	if (!err == -1 ) CUDA_CHECK_CALL(CUDA_ERROR_FILE_NOT_FOUND);
//	if (!err == -2 ) CUDA_CHECK_CALL(CUDA_ERROR_UNKNOWN);
	// in this branch we use compilation with parameters

	// set up size of compilation log buffer
	jitOptions[0] = CU_JIT_INFO_LOG_BUFFER_SIZE_BYTES;
	jitLogBufferSize = 1024;
	jitOptVals[0] = (void *)(size_t)jitLogBufferSize;

	// set up pointer to the compilation log buffer
	jitOptions[1] = CU_JIT_INFO_LOG_BUFFER;
	jitLogBuffer = malloc(sizeof(char)*jitLogBufferSize);
	jitOptVals[1] = jitLogBuffer;

	// set up pointer to set the Maximum # of registers for a particular kernel
	jitOptions[2] = CU_JIT_MAX_REGISTERS;

	jitOptVals[2] = (void *)(size_t)jitRegCount;
	CUDA_CHECK_CALL(cuModuleLoadDataEx(&module, ptx_source, jitNumOptions, jitOptions, (void **)jitOptVals));
	//CUDA_CHECK_CALL(cuModuleLoad(&module, ptx_fname));
	CUDA_CHECK_CALL(cuModuleGetFunction(kernel, module, functionN));
	CUDA_RESTORE_CONTEXT;
	CAMLreturn((value) kernel);
}

CAMLprim value spoc_cuda_create_extra(value n){
	CAMLparam1(n);
	char *extra = malloc(sizeof(void*)*Int_val(n)*2);
	CAMLreturn ((value)extra);
}

CAMLprim value spoc_cuda_create_dummy_kernel(){
	CAMLparam0();
	CUfunction *kernel;
	CAMLreturn((value) kernel);
}

#define ALIGN_UP(offset, alignment) \
(offset) = ((offset) + (alignment) - 1) & ~((alignment) - 1)


#define ADD_TO_PARAM_BUFFER(value, alignment)\
do {\
	offset = ALIGN_UP(offset, alignment); \
memcpy(extra + offset, &(value), sizeof(value));\
offset += sizeof(value);\
} while (0)



CAMLprim value spoc_cuda_load_param_vec(value off, value ex, value A){
	CAMLparam3(off, ex, A);
	CUdeviceptr d_A;
	char *extra;
	int offset;
	extra = (char*)ex;
	offset = Int_val(Field(off, 0));
	d_A = CUdeviceptr_val(Field(A, 1));

	ADD_TO_PARAM_BUFFER(d_A, __alignof(d_A));
	Store_field(off, 0, Val_int(offset));

	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cuda_custom_load_param_vec(value off, value ex, value A){
	CAMLparam3(off, ex, A);
	CUdeviceptr d_A;
	char *extra;
	int offset;
	extra = (char*)ex;
	offset = Int_val(Field(off, 0));
	d_A = CUdeviceptr_val((Field(A, 1)));

	ADD_TO_PARAM_BUFFER(d_A, __alignof(d_A));
	Store_field(off, 0, Val_int(offset));

	CAMLreturn(Val_unit);
}


CAMLprim value spoc_cuda_load_param_int(value off, value ex,value val){
	CAMLparam3(off, ex, val);
	int offset, i;// = malloc(sizeof(i));
	char*extra;
	extra = (char*)ex;
	offset = Int_val(Field(off, 0));
	i = Int_val(val);
	ADD_TO_PARAM_BUFFER(i, __alignof(i));
	Store_field(off, 0, Val_int(offset));
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cuda_load_param_int64(value off, value ex, value val){
	CAMLparam3(off, ex, val);
	int offset;
	char *extra;
	long long i;
	extra = (char*)ex;
	offset = Int_val(Field(off, 0));
	i = (long long) Int_val(val);
	ADD_TO_PARAM_BUFFER(i, __alignof(i));
	Store_field(off, 0, Val_int(offset));
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cuda_load_param_float(value off, value ex, value val){
	CAMLparam3(off, ex, val);
	int offset;
	char *extra;
	float f;
	offset = Int_val(Field(off, 0));
	extra = (char*)ex;
	f = (float)Double_val(val);
	ADD_TO_PARAM_BUFFER(f, __alignof(f));
	Store_field(off, 0, Val_int(offset));
	CAMLreturn(Val_unit);
}
CAMLprim value spoc_cuda_load_param_float64(value off, value ex, value val){
	CAMLparam3(off, ex, val);
	int offset;
	char *extra;
	double f;
	extra = (char*)ex;
	offset = Int_val(Field(off, 0));
	f = Double_val(val);
	ADD_TO_PARAM_BUFFER(f, __alignof(f));
	Store_field(off, 0, Val_int(offset));
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cuda_load_param_vec_b(value off, value ex, value A){
	return spoc_cuda_load_param_vec(off, ex, A);

}
CAMLprim value spoc_cuda_load_param_vec_n(value off, value ex, value A){
	return spoc_cuda_load_param_vec(off, ex, A);
}

CAMLprim value spoc_cuda_custom_load_param_vec_b(value off, value ex, value A){
	return spoc_cuda_custom_load_param_vec(off, ex, A);

}
CAMLprim value spoc_cuda_custom_load_param_vec_n(value off, value ex, value A){
	return spoc_cuda_custom_load_param_vec(off, ex, A);
}

CAMLprim value spoc_cuda_load_param_int_b(value off, value ex, value val){
	return spoc_cuda_load_param_int(off, ex, val);

}
CAMLprim value spoc_cuda_load_param_int_n(value off, value ex, value val){
	return spoc_cuda_load_param_int(off, ex, val);
}


CAMLprim value spoc_cuda_load_param_int64_b(value off, value ex, value val){
	return spoc_cuda_load_param_int64(off, ex, val);

}
CAMLprim value spoc_cuda_load_param_int64_n(value off, value ex, value val){
	return spoc_cuda_load_param_int64(off, ex, val);
}


CAMLprim value spoc_cuda_load_param_float_b(value off, value ex, value val){
	return spoc_cuda_load_param_float(off, ex, val);

}
CAMLprim value spoc_cuda_load_param_float_n(value off, value ex, value val){
	return spoc_cuda_load_param_float(off, ex, val);
}


CAMLprim value spoc_cuda_load_param_float64_b(value off, value ex, value val){
	return spoc_cuda_load_param_float64(off, ex, val);

}
CAMLprim value spoc_cuda_load_param_float64_n(value off, value ex, value val){
	return spoc_cuda_load_param_float64(off, ex, val);
}

CAMLprim value spoc_cuda_set_block_shape(value ker, value block, value gi){
	CAMLparam3(ker, block, gi);
	CUfunction *kernel;

	CUDA_GET_CONTEXT;

	kernel = (CUfunction*) ker;
	CUDA_CHECK_CALL(cuFuncSetBlockShape(*kernel, Int_val(Field(block,0)),Int_val(Field(block,1)),Int_val(Field(block,2))));

	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cuda_launch_grid(value off, value ker, value grid, value block,  value ex, value gi){
	CAMLparam5(ker, grid, ex, block, gi);
	CUfunction *kernel;
	int gridX, gridY, gridZ, blockX, blockY, blockZ;
	int offset;
	char* extra;
	void* extra2[5];
	CUDA_GET_CONTEXT;

	offset = Int_val(Field(off, 0));
	kernel = (CUfunction*) ker;
	gridX = Int_val(Field(grid,0));
	gridY = Int_val(Field(grid,1));
	gridZ = Int_val(Field(grid,2));

	blockX = Int_val(Field(block,0));
	blockY = Int_val(Field(block,1));
	blockZ = Int_val(Field(block,2));

	extra = (char*)ex;

	extra2[0] =	CU_LAUNCH_PARAM_BUFFER_POINTER;
	extra2[1] = extra;
	extra2[2] = CU_LAUNCH_PARAM_BUFFER_SIZE;
	extra2[3] = &offset;
	extra2[4] = CU_LAUNCH_PARAM_END;


	CUDA_CHECK_CALL(cuLaunchKernel(*kernel, gridX, gridY, gridZ, blockX, blockY, blockZ, 0, 0, NULL, extra2));
	//CUDA_CHECK_CALL(cuCtxSynchronize());
	Store_field(off, 0, Val_int(offset));
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cuda_flush(value gi){
	CAMLparam1(gi);
	CUDA_GET_CONTEXT;
	cuCtxSynchronize ();
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cuda_launch_grid_b(value off, value ker, value grid, value block,  value ex, value gi){
	return spoc_cuda_launch_grid(off, ker, grid, block,  ex, gi);
}
CAMLprim value spoc_cuda_launch_grid_n(value off, value ker, value grid, value block,  value ex, value gi){
	return spoc_cuda_launch_grid(off, ker, grid, block, ex, gi);
}
#ifdef __cplusplus
}
#endif
