#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/custom.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <caml/bigarray.h>
#include <math.h>
#include <string.h>
#include "Spoc.h"


CAMLprim value spoc_opencl_load(value moduleSrc, value function_name, value gi){
	CAMLparam3(moduleSrc, function_name, gi);
	cl_program hProgram;
	cl_device_id device_id;
	cl_kernel kernel;
	char* functionN;
	char* cl_source;
//	int err;

	OPENCL_GET_CONTEXT;
//	moduleN = String_val(module_name);
//	functionN = String_val(function_name);
//	cl_fname = malloc(sizeof(char)*(strlen(moduleN)+3));
//	strcpy (cl_fname, moduleN);
//	strcat(cl_fname, ".cl");
//	moduleN = String_val(module_name);
	functionN = String_val(function_name);
	cl_source = String_val(moduleSrc);
//	err = ae_load_file_to_memory(cl_fname, &cl_source);
//	if (err == -1) OPENCL_CHECK_CALL1(opencl_error, CL_FILE_NOT_FOUND);
	cl_int ret_val;
	OPENCL_CHECK_CALL1(hProgram, clCreateProgramWithSource(ctx, 1, (const char**)&cl_source, 0, &opencl_error));
	OPENCL_TRY("clGetContextInfo", clGetContextInfo(ctx, CL_CONTEXT_DEVICES, (size_t)sizeof(cl_device_id), &device_id, NULL)) ;
	OPENCL_CHECK_CALL1(ret_val, clBuildProgram(hProgram, 1, &device_id, 0, NULL, NULL));

	size_t paramValueSize = 1024 * 1024,
	         param_value_size_ret;
	   char *paramValue;
	   paramValue = (char*)calloc(paramValueSize, sizeof(char));
	   ret_val = clGetProgramBuildInfo( hProgram,
			   device_id,
	                           CL_PROGRAM_BUILD_LOG,
	                           paramValueSize,
	                           paramValue,
	                           &param_value_size_ret);

	//   fprintf(stdout, " %s" , paramValue);

	// create kernel
	OPENCL_CHECK_CALL1(kernel, clCreateKernel(hProgram, functionN, &opencl_error));
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn((value) kernel);;

}


CAMLprim value spoc_debug_opencl_load(value moduleSrc, value function_name, value gi){
	CAMLparam3(moduleSrc, function_name, gi);
	cl_program hProgram;
	cl_device_id device_id;
	cl_kernel kernel;
	char* functionN;
	char* cl_source;
//	int err;

	OPENCL_GET_CONTEXT;
//	moduleN = String_val(module_name);
//	functionN = String_val(function_name);
//	cl_fname = malloc(sizeof(char)*(strlen(moduleN)+3));
//	strcpy (cl_fname, moduleN);
//	strcat(cl_fname, ".cl");
//	moduleN = String_val(module_name);
	functionN = String_val(function_name);
	cl_source = String_val(moduleSrc);
//	err = ae_load_file_to_memory(cl_fname, &cl_source);
//	if (err == -1) OPENCL_CHECK_CALL1(opencl_error, CL_FILE_NOT_FOUND);
	cl_int ret_val;
	OPENCL_CHECK_CALL1(hProgram, clCreateProgramWithSource(ctx, 1, (const char**)&cl_source, 0, &opencl_error));
	OPENCL_TRY("clGetContextInfo", clGetContextInfo(ctx, CL_CONTEXT_DEVICES, (size_t)sizeof(cl_device_id), &device_id, NULL)) ;
	OPENCL_CHECK_CALL1(ret_val, clBuildProgram(hProgram, 1, &device_id, 0, NULL, NULL));

	size_t paramValueSize = 1024 * 1024,
	         param_value_size_ret;
	   char *paramValue;
	   paramValue = (char*)calloc(paramValueSize, sizeof(char));
	   ret_val = clGetProgramBuildInfo( hProgram,
			   device_id,
	                           CL_PROGRAM_BUILD_LOG,
	                           paramValueSize,
	                           paramValue,
	                           &param_value_size_ret);

	   fprintf(stdout, " %s" , paramValue);

	// create kernel
	OPENCL_CHECK_CALL1(kernel, clCreateKernel(hProgram, functionN, &opencl_error));
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn((value) kernel);;

}


CAMLprim value spoc_opencl_create_dummy_kernel(){
	CAMLparam0();
	cl_kernel kernel;
	CAMLreturn((value) kernel);
}

CAMLprim value spoc_opencl_load_param_vec(value off, value ker, int idx, value A, value gi){
	CAMLparam4(off, ker, A, gi);
	cl_kernel kernel;
	cl_mem d_A;
	int offset;
	OPENCL_GET_CONTEXT;
	offset = Int_val(Field(off, 0));
	d_A = Cl_mem_val(Field(A, 1));
	kernel = (cl_kernel) ker;
	OPENCL_CHECK_CALL1(opencl_error, clSetKernelArg(kernel, offset, sizeof(cl_mem), (void*)&d_A));
	offset+=1;
	Store_field(off, 0, Val_int(offset));
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_opencl_load_param_local_vec(value off, value ker, int idx, value A, value gi){
	CAMLparam4(off, ker, A, gi);
	cl_kernel kernel;
	cl_mem d_A;
	int offset;
	OPENCL_GET_CONTEXT;
	offset = Int_val(Field(off, 0));
	d_A = Cl_mem_val(Field(A, 1));
	kernel = (cl_kernel) ker;
	OPENCL_CHECK_CALL1(opencl_error, clSetKernelArg(kernel, offset, sizeof(float)*4*8*8, NULL));
	offset+=1;
	Store_field(off, 0, Val_int(offset));
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_opencl_load_param_int(value off, value ker, value val, value gi){
	CAMLparam4(off, ker, val, gi);
	cl_kernel kernel;
	int i;
	int offset;
	OPENCL_GET_CONTEXT;
	offset = Int_val(Field(off, 0));
	i = Int_val(val);
	kernel = (cl_kernel) ker;
	OPENCL_CHECK_CALL1(opencl_error, clSetKernelArg(kernel, offset, sizeof(int), (void*)&i));
	offset+=1;
	Store_field(off, 0, Val_int(offset));
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_opencl_load_param_int64(value off, value ker, value val, value gi){
	CAMLparam4(off, ker, val, gi);
	cl_kernel kernel;
	long long i;
	int offset;
	OPENCL_GET_CONTEXT;
	offset = Int_val(Field(off, 0));

	i = (long long)Int_val(val);
	kernel = (cl_kernel) ker;
	OPENCL_CHECK_CALL1(opencl_error, clSetKernelArg(kernel, offset, sizeof(long long), (void*)&i));
	offset+=1;
	Store_field(off, 0, Val_int(offset));
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_opencl_load_param_float(value off, value ker, value val, value gi){
	CAMLparam4(off, ker, val, gi);
	cl_kernel kernel;
	float f;
	int offset;
	OPENCL_GET_CONTEXT;
	offset = Int_val(Field(off, 0));

	f = (float)(Double_val(val));
	kernel = (cl_kernel) ker;
	OPENCL_CHECK_CALL1(opencl_error, clSetKernelArg(kernel, offset, sizeof(float), (void*)&f));
	offset+=1;
	Store_field(off, 0, Val_int(offset));
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_opencl_load_param_float64(value off, value ker, value val, value gi){
	CAMLparam4(off, ker, val, gi);
	cl_kernel kernel;
	double f;
	int offset;
	OPENCL_GET_CONTEXT;
	offset = Int_val(Field(off, 0));

	f = (Double_val(val));
	kernel = (cl_kernel) ker;
	OPENCL_CHECK_CALL1(opencl_error, clSetKernelArg(kernel, offset, sizeof(double), (void*)&f));
	offset+=1;
	Store_field(off, 0, Val_int(offset));
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_opencl_launch_grid(value ker, value grid, value block, value gi, value queue_id){
	CAMLparam5(ker, grid, block, gi, queue_id);
	cl_kernel kernel;
	unsigned int gridX, gridY, gridZ, blockX, blockY, blockZ;
	unsigned int dimension;
	OPENCL_GET_CONTEXT;

	kernel = (cl_kernel) ker;
	gridX = Int_val(Field(grid,0));
	gridY = Int_val(Field(grid,1));
	gridZ = Int_val(Field(grid,2));

	blockX = Int_val(Field(block,0));
	blockY = Int_val(Field(block,1));
	blockZ = Int_val(Field(block,2));
	size_t *work_size;
	size_t global_dimension[3] = {(size_t)gridX*blockX,(size_t)gridY*blockY,(size_t)gridZ*blockZ};
	if ((blockX == 1) && (blockY == 1) && (blockZ == 1))
		work_size = (size_t*)NULL;
	else
	{
			work_size = malloc (sizeof(size_t)*3);
			work_size[0] = (size_t)blockX;
			work_size[1] = (size_t)blockY;
			work_size[2] = (size_t)blockZ;
	}
	//OPENCL_CHECK_CALL1(opencl_error, clFinish(queue[Int_val(queue_id)]));
   	cl_command_queue q = queue[Int_val(queue_id)];
   	//printf("create %p\n", q); fflush(stdout);
	OPENCL_CHECK_CALL1(opencl_error, clRetainCommandQueue(queue[Int_val(queue_id)]));
	OPENCL_CHECK_CALL1(opencl_error, clEnqueueNDRangeKernel(q, kernel, 3, NULL, (size_t*) global_dimension, (size_t*)work_size, 0, NULL, NULL));
	//OPENCL_CHECK_CALL1(opencl_error, clFinish(q));
	OPENCL_CHECK_CALL1(opencl_error, clReleaseCommandQueue(queue[Int_val(queue_id)]));

	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}


CAMLprim value spoc_opencl_flush(value gi, value queue_id){
	CAMLparam2(gi, queue_id);
	OPENCL_GET_CONTEXT;
	//printf("spoc_opencl_flush %p\n",queue[Int_val(queue_id)]); fflush(stdout);
   	cl_command_queue q = queue[Int_val(queue_id)];
	OPENCL_CHECK_CALL1(opencl_error, clFinish(q));
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}


#ifdef __cplusplus
}
#endif
