open Vector
  
open Devices
  
open Cuda
  
open OpenCL
  
let auto = ref true
  
let auto_transfers b = auto := b
  
  
let to_device vect ?queue_id:(q = 0) (dev : device) =
  if vect.dev <> (Dev dev)
  then
    (match dev.general_info.specific with
     | Cuda ->
         (match vect.vector with
          | Vector.CustomArray custom ->
              (cuda_custom_cpu_to_device vect dev.general_info.id
                 dev.general_info;
               vect.device <- dev.general_info.id;
							vect.dev <- Dev dev)
          | _ ->
              (current_cuda_device := dev.general_info.id;
               match vect.dev with
								| No_dev  -> 
                 (try
                    cuda_cpu_to_device vect dev.general_info.id
                      dev.general_info;
											vect.device <- dev.general_info.id;
											vect.dev <- Dev dev
                  with
                  | Cuda.ERROR_OUT_OF_MEMORY ->
                      (Gc.full_major ();
                       cuda_cpu_to_device vect dev.general_info.id
                         dev.general_info;
												vect.device <- dev.general_info.id;
												vect.dev <- Dev dev);
												)
                  | Dev d ->
                      (match d.general_info.specific with
                       | Cuda ->
                           cuda_device_to_device vect dev.general_info.id d;
													vect.device <- dev.general_info.id;
													vect.dev <- Dev dev
                       | OpenCL ->
                           (opencl_device_to_cpu vect d.general_info.id
                              d.general_info d.specific_info q;
                            (try
                               cuda_cpu_to_device vect dev.general_info.id
                                 dev.general_info;
																vect.device <- dev.general_info.id;
																vect.dev <- Dev dev
                             with
                             | Cuda.ERROR_OUT_OF_MEMORY ->
                                 (Gc.full_major ();
                                  cuda_cpu_to_device vect dev.general_info.id
                                    dev.general_info;
																		vect.device <- dev.general_info.id;
																		vect.dev <- Dev dev);
																		))
                       | _ -> failwith "CU TO_Device Not Implemented");
               vect.device <- dev.general_info.id;
               vect.dev <- Dev dev))
     | OpenCL ->
         (match vect.vector with
          | Vector.CustomArray custom -> 
						(opencl_custom_cpu_to_device vect (dev.general_info.id - !cuda_compatible_devices)
                 dev.general_info q;
               vect.device <- dev.general_info.id;
							vect.dev <- Dev dev)
          | _ ->
              (current_cuda_device := dev.general_info.id;
               match vect.dev with
								| No_dev -> 
                 (try
                    opencl_cpu_to_device vect
                      (dev.general_info.id - !cuda_compatible_devices)
                      dev.general_info q;
											vect.device <- dev.general_info.id;
											vect.dev <- Dev dev
                  with
                  | OpenCL.MEM_OBJECT_ALLOCATION_FAILURE ->
                      (Gc.full_major ();
                       opencl_cpu_to_device vect
                         (dev.general_info.id - !cuda_compatible_devices)
                         dev.general_info q;
												vect.device <- dev.general_info.id;
												vect.dev <- Dev dev);
												)
                  | Dev d ->
                      (match d.general_info.specific with
                       | Cuda ->
                           (cuda_device_to_cpu vect d.general_info.id
                              d.general_info;
															vect.device <- dev.general_info.id;
															vect.dev <- Dev dev;
                            (try
                               opencl_cpu_to_device vect
                                 (dev.general_info.id -
                                    !cuda_compatible_devices)
                                 dev.general_info q;
																vect.device <- dev.general_info.id;
																vect.dev <- Dev dev
                             with
                             | OpenCL.MEM_OBJECT_ALLOCATION_FAILURE ->
                                 (Gc.full_major ();
                                  opencl_cpu_to_device vect
                                    (dev.general_info.id -
                                       !cuda_compatible_devices)
                                    dev.general_info q;
																		vect.device <- dev.general_info.id;
																		vect.dev <- Dev dev);
																		))
                       | OpenCL ->
                           opencl_device_to_device vect (d.general_info.id -
                                       !cuda_compatible_devices) d
                       | _ -> failwith "CL TO_Device Not Implemented");
               vect.device <- dev.general_info.id;
               vect.dev <- Dev dev))
     | _ -> failwith "TO_Device Not Implemented")
  else ()
  
let to_cpu vect ?queue_id:(q= 0) () =
  ( match vect.dev with
  | Dev d ->
      (match d.general_info.specific with
       | Cuda ->
           (match vect.vector with
            | Vector.CustomArray custom ->
                (cuda_custom_device_to_cpu vect d.general_info.id
                   d.general_info;
                 vect.device <- (-1);
								 vect.dev <- No_dev)
            | _ ->
                (cuda_device_to_cpu vect d.general_info.id d.general_info;
                 vect.device <- (-1);
                 vect.dev <- No_dev))
       | OpenCL ->
				(match vect.vector with
					| Vector.CustomArray custom ->
								(opencl_custom_device_to_cpu vect (d.general_info.id - !cuda_compatible_devices) d.general_info  d.specific_info q;
                 vect.device <- (-1);
								 vect.dev <- No_dev)
					| _  -> 
           (opencl_device_to_cpu vect
              (d.general_info.id - !cuda_compatible_devices) d.general_info
              d.specific_info q;
            vect.device <- (-1);
            vect.dev <- No_dev))
       | _ -> failwith "TO_CPU Not Implemented")
  | No_dev -> ()
	)
  
let unsafe_set vect idx value =
  match vect.vector with
  | Bigarray v -> Bigarray.Array1.set v idx value
  | CustomArray c -> (snd c).set (fst c) idx value
  
let unsafe_get vect idx =
  match vect.vector with
  | Bigarray v -> Bigarray.Array1.get v idx
  | CustomArray c -> (snd c).get (fst c) idx
  
(* external set: ('a,'b,'c) Bigarray.Array1.t -> int -> 'a -> unit =       *)
(* "spoc_set"                                                              *)
let set vect (idx : int) (value : 'a) =
  if idx >= vect.length
  then failwith "index out of bounds"
  else
    (if !auto
     then
       (match vect.dev with
        | Dev d ->
              (Devices.flush d ~queue_id: 0 ();
               Devices.flush d ~queue_id: 1 ();
               to_cpu vect ();
               Devices.flush d ~queue_id: 0 ())
        | No_dev -> ())
     else ();
     unsafe_set vect idx value)
  
(* external get: ('a,'b,'c) Bigarray.Array1.t -> int -> 'a = "spoc_get" *)
let get vect idx =
  if idx >= vect.length
  then failwith "index out of bounds"
  else
    (if !auto
     then
       (match vect.dev with
        | Dev d ->
              (Devices.flush d ~queue_id: 0 ();
               to_cpu vect ~queue_id: 0 ();
               Devices.flush d ~queue_id: 0 ())
        | No_dev -> ())
     else ();
     unsafe_get vect idx)
  
external sub_custom_array : customarray -> 'a custom -> int -> customarray =
  "spoc_sub_custom_array"
  
let sub_vector (vect : ('a, 'b) vector) _start _len =
  if (_start + _len) > vect.length
  then failwith "index out of bounds"
  else
    ((match vect.dev with
      | Dev d ->
            (Devices.flush d ~queue_id: 0 ();
             to_cpu vect ~queue_id: 0 ();
             Devices.flush d ~queue_id: 0 ();)
      | No_dev -> ());
     (match vect.vector with
      | Bigarray b ->
          {
            device = (-1);
            vector = Bigarray (Bigarray.Array1.sub b _start _len);
            cuda_device_vec =
              Array.create (!cuda_compatible_devices + 1)
                (init_cuda_device_vec ());
            opencl_device_vec =
              Array.create (!opencl_compatible_devices + 1)
                (init_opencl_device_vec ());
            length = _len;
            dev = No_dev;
						kind = vect.kind;
          }
      | CustomArray (cA, c) ->
          {
            device = (-1);
            vector = CustomArray ((sub_custom_array cA c _start), c);
            cuda_device_vec =
              Array.create (!cuda_compatible_devices + 1)
                (init_cuda_device_vec ());
            opencl_device_vec =
              Array.create (!opencl_compatible_devices + 1)
                (init_opencl_device_vec ());
            length = _len;
            dev = No_dev;
						kind = vect.kind;
          }))
  
