#ifdef __cplusplus
extern "C" {
#endif

#include "Mem_c.h"
#include <assert.h>

void cuda_free_vec (value v) {
	CUdeviceptr f = CUdeviceptr_val(Field(v, 1));
	if (f) cuMemFree(f);
}


CAMLprim value spoc_init_cuda_device_vec(){
	CAMLparam0();
	CAMLlocal1(ret);
	CUdeviceptr f = 0;
	ret = alloc_final(2, cuda_free_vec, sizeof(CUdeviceptr), sizeof(CUdeviceptr)*1024);
	Store_field(ret, 1, Val_CUdeviceptr(f));
	CAMLreturn(ret);
}

void cl_free_vec (value v) {
	cl_mem f = Cl_mem_val(Field(v, 1));
	if (f) clReleaseMemObject(f);
}

CAMLprim value spoc_init_opencl_device_vec(){
	CAMLparam0();
	CAMLlocal1(ret);
	cl_mem f = NULL;
	ret = alloc_final(2, cl_free_vec, sizeof(cl_mem), sizeof(cl_mem)*1024);
	Store_field(ret, 1, Val_cl_mem(f));

	CAMLreturn(ret);
}



CAMLprim value spoc_cuda_cpu_to_device(value vector, value nb_device, value gi){
	CAMLparam3(vector, nb_device, gi);
	CAMLlocal5(device_vec, vec, bigArray,  dev_vec_array, dev_vec);
	void* h_A;
	CUdeviceptr d_A;
	int size;
	int type_size;
	int tag;
	CUDA_GET_CONTEXT;
	//vec = Field (vector, 1);
	bigArray = Field (Field(vector, 1), 0);
	h_A = (void*)Data_bigarray_val(bigArray);
	GET_TYPE_SIZE;
	dev_vec_array = Field(vector, 2);
	dev_vec =Field(dev_vec_array, Int_val(nb_device));
	d_A = CUdeviceptr_val(Field(dev_vec, 1));
	if (&d_A != NULL) cuMemFree(d_A);
	size =Int_val(Field(vector, 4));
	CUDA_CHECK_CALL(cuMemAlloc(&d_A, size*type_size));
	CUDA_CHECK_CALL(cuMemcpyHtoD(d_A, h_A, size*type_size));
	//CUDA_CHECK_CALL(cuCtxSynchronize());
	Store_field(dev_vec, 1, Val_CUdeviceptr(d_A));
	Store_field(dev_vec_array, Int_val(nb_device), dev_vec);
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cuda_custom_cpu_to_device(value vector, value nb_device, value gi){
	CAMLparam3(vector, nb_device, gi);
	CAMLlocal5(device_vec, vec, customArray,  dev_vec_array, dev_vec);
	char* h_A;
	CUdeviceptr d_A;
	int size;
	int type_size;
	int tag;
	CUDA_GET_CONTEXT;
	//vec = Field (vector, 1);
	customArray = Field (Field(vector, 1), 0);
	h_A = (char*)Field(Field(customArray, 0),1);

	//GET_TYPE_SIZE;
	type_size = Int_val(Field(Field(customArray, 1),1));
	size =Int_val(Field(vector, 4));

	dev_vec_array = Field(vector, 2);
	dev_vec =Field(dev_vec_array, Int_val(nb_device));
	d_A = CUdeviceptr_val(Field(dev_vec, 1));

	if (NULL != &d_A) cuMemFree(d_A);
	CUDA_CHECK_CALL(cuMemAlloc(&d_A, size*type_size));
	CUDA_CHECK_CALL(cuMemcpyHtoD(d_A, h_A, size*type_size));

	//CUDA_CHECK_CALL(cuCtxSynchronize());

	Store_field(dev_vec, 1, Val_CUdeviceptr(d_A));
	Store_field(dev_vec_array, Int_val(nb_device), dev_vec);
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}



CAMLprim value spoc_cuda_device_to_device() {
	CAMLparam0();
	UNIMPLENTED;

	CAMLreturn(Val_unit);

}

CAMLprim value spoc_cuda_device_to_cpu(value vector, value nb_device, value gi) {
	CAMLparam3(vector, nb_device, gi);
	CAMLlocal5(device_vec, vec, bigArray,  dev_vec_array, dev_vec);
	CUdevice dev;
	void* h_A;
	CUdeviceptr d_A;
	int id;
	int size;
	int type_size;
	int tag;
	CUDA_GET_CONTEXT;


	CUDA_CHECK_CALL(cuDeviceGet(&dev, Int_val(nb_device)));
	//vec = Field (vector, 1);
	id = Int_val(Field(vector, 0));

	bigArray = Field (Field(vector, 1), 0);
	h_A = (void*)Data_bigarray_val(bigArray);
	GET_TYPE_SIZE;

	dev_vec_array = Field(vector, 2);
	dev_vec =Field(dev_vec_array, Int_val(nb_device));
	d_A = CUdeviceptr_val(Field(dev_vec, 1));

	size = Int_val(Field(vector, 4));
	CUDA_CHECK_CALL(cuMemcpyDtoH((void*)h_A, d_A, size*type_size));
	CUDA_RESTORE_CONTEXT;

	CAMLreturn(Val_unit);
}


CAMLprim value spoc_cuda_custom_device_to_cpu(value vector, value nb_device, value gi) {
	CAMLparam3(vector, nb_device, gi);
	CAMLlocal5(device_vec, vec, customArray,  dev_vec_array, dev_vec);
	CUdevice dev;
	void* h_A;
	CUdeviceptr d_A;
	int id;
	int size;
	int type_size;
	int tag;
	CUDA_GET_CONTEXT;


	CUDA_CHECK_CALL(cuDeviceGet(&dev, Int_val(nb_device)));
	//vec = Field (vector, 1);
	id = Int_val(Field(vector, 0));
	customArray = Field (Field(vector, 1), 0);
	h_A = (void*)Field(Field(customArray, 0), 1);
	//GET_TYPE_SIZE;
	type_size = Int_val(Field(Field(customArray, 1),1))*sizeof(char);
	size = Int_val(Field(vector, 4));
	dev_vec_array = Field(vector, 2);
	dev_vec =Field(dev_vec_array, Int_val(nb_device));
	d_A = CUdeviceptr_val(Field(dev_vec, 1));

	CUDA_CHECK_CALL(cuMemcpyDtoH((void*)h_A, d_A, size*type_size));
	CUDA_RESTORE_CONTEXT;

	CAMLreturn(Val_unit);
}



value spoc_opencl_cpu_to_device(value vector, value nb_device, value gi, value queue_id) {
	CAMLparam4(vector, nb_device, gi, queue_id);
	CAMLlocal5(device_vec, vec, bigArray,  dev_vec_array, dev_vec);
	void* h_A;
	cl_mem d_A;
	int size;
	cl_device_id device_id;
	int type_size;
	int tag;
	OPENCL_GET_CONTEXT;

	//vec = Field (vector, 1);
	bigArray = Field (Field(vector, 1), 0);
	h_A = (void*)Data_bigarray_val(bigArray);

	GET_TYPE_SIZE

	//h_A = (void*)Data_bigarray_val(bigArray);
	dev_vec_array = Field(vector, 3);
	dev_vec =Field(dev_vec_array, Int_val(nb_device));
	d_A = Cl_mem_val(Field(dev_vec, 1));

	if (NULL != d_A) clReleaseMemObject(d_A);
	size =Int_val(Field(vector, 4));
	OPENCL_CHECK_CALL1(d_A, clCreateBuffer(ctx, CL_MEM_READ_WRITE, size*type_size, NULL, &opencl_error));
	//clRetainMemObject(d_A);
	//OPENCL_TRY("clGetContextInfo", clGetContextInfo(ctx, CL_CONTEXT_DEVICES, (size_t)sizeof(cl_device_id), &device_id, NULL)) ;
	//OPENCL_CHECK_CALL1(opencl_error, clRetainCommandQueue(queue[Int_val(queue_id)]));
	OPENCL_CHECK_CALL1(opencl_error, clEnqueueWriteBuffer(queue[Int_val(queue_id)], d_A, CL_TRUE, 0, size*type_size, h_A, 0, NULL, NULL));
	//OPENCL_CHECK_CALL1(opencl_error, clReleaseCommandQueue(queue[Int_val(queue_id)]));
	Store_field(dev_vec, 1, Val_cl_mem(d_A));
	//printf ("to_device : %p\n", d_A); fflush(stdout);
	Store_field(dev_vec_array, Int_val(nb_device), dev_vec);
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

value spoc_opencl_custom_cpu_to_device(value vector, value nb_device, value gi, value queue_id) {
	CAMLparam4(vector, nb_device, gi, queue_id);
	CAMLlocal5(device_vec, vec, customArray,  dev_vec_array, dev_vec);
	void* h_A;
	cl_mem d_A;
	int size;
	cl_device_id device_id;
	int type_size;
	int tag;
	OPENCL_GET_CONTEXT;

	//vec = Field (vector, 1);
	customArray = Field (Field(vector, 1), 0);
	h_A = (void*)Field(Field(customArray, 0), 1);

	//GET_TYPE_SIZE
	type_size = Int_val(Field(Field(customArray, 1),1))*sizeof(char);
	size = Int_val(Field(vector, 4));

	//h_A = (void*)Data_bigarray_val(bigArray);
	dev_vec_array = Field(vector, 3);
	dev_vec =Field(dev_vec_array, Int_val(nb_device));
	d_A = Cl_mem_val(Field(dev_vec, 1));
	//if (NULL != d_A) clReleaseMemObject(d_A);
	OPENCL_CHECK_CALL1(d_A, clCreateBuffer(ctx, CL_MEM_READ_WRITE, size*type_size, NULL, &opencl_error));
	OPENCL_TRY("clGetContextInfo", clGetContextInfo(ctx, CL_CONTEXT_DEVICES, (size_t)sizeof(cl_device_id), &device_id, NULL)) ;
	//OPENCL_CHECK_CALL1(opencl_error, clRetainCommandQueue(queue[Int_val(queue_id)]));
	OPENCL_CHECK_CALL1(opencl_error, clEnqueueWriteBuffer(queue[Int_val(queue_id)], d_A, CL_TRUE, 0, size*type_size, h_A, 0, NULL, NULL));
	//OPENCL_CHECK_CALL1(opencl_error, clReleaseCommandQueue(queue[Int_val(queue_id)]));

	Store_field(dev_vec, 1, Val_cl_mem(d_A));

	Store_field(dev_vec_array, Int_val(nb_device), dev_vec);
	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}


value spoc_opencl_device_to_device() {
	CAMLparam0();
	UNIMPLENTED;
	fflush(stdout);

	CAMLreturn(Val_unit);

}

value spoc_opencl_device_to_cpu(value vector, value nb_device, value gi, value si, value queue_id) {
	CAMLparam5(vector, nb_device, gi, si, queue_id);
	CAMLlocal5(device_vec, vec, bigArray,  dev_vec_array, dev_vec);
	void* h_A;
	cl_mem d_A;
	int size;
	int type_size;
	int tag;
	cl_device_id device_id;
	OPENCL_GET_CONTEXT;

//	vec = Field (vector, 1);
	bigArray = Field (Field(vector, 1), 0);
	h_A = (void*)Data_bigarray_val(bigArray);

	GET_TYPE_SIZE


	//h_A = (void*)Data_bigarray_val(bigArray);
	dev_vec_array = Field(vector, 3);
	dev_vec =Field(dev_vec_array, Int_val(nb_device));
	d_A = Cl_mem_val(Field(dev_vec, 1));

	size = Int_val(Field(vector, 4));

	cl_command_queue  q = queue[Int_val(queue_id)];

	//OPENCL_TRY("clGetContextInfo", clGetContextInfo(ctx, CL_CONTEXT_DEVICES, (size_t)sizeof(cl_device_id), &device_id, NULL)) ;
	//OPENCL_CHECK_CALL1(opencl_error, clRetainCommandQueue(queue[Int_val(queue_id)]));
	OPENCL_CHECK_CALL1(opencl_error, clEnqueueReadBuffer(q, d_A, CL_TRUE, 0, size*type_size, h_A, 0, NULL, NULL));
	//OPENCL_CHECK_CALL1(opencl_error, clReleaseCommandQueue(queue[Int_val(queue_id)]));
	//OPENCL_CHECK_CALL1(opencl_error, clFlush(queue[Int_val(queue_id)]));

	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

value spoc_opencl_custom_device_to_cpu(value vector, value nb_device, value gi, value si, value queue_id) {
	CAMLparam5(vector, nb_device, gi, si, queue_id);
	CAMLlocal5(device_vec, vec, customArray,  dev_vec_array, dev_vec);
	void* h_A;
	cl_mem d_A;
	int size;
	int type_size;
	int tag;
	cl_device_id device_id;
	OPENCL_GET_CONTEXT;

//	vec = Field (vector, 1);
	customArray = Field (Field(vector, 1), 0);
	h_A = (void*)Field(customArray, 0);
	//GET_TYPE_SIZE;

	//h_A = (void*)Data_bigarray_val(bigArray);
	dev_vec_array = (value)Field(vector, 3);
	dev_vec =Field(dev_vec_array, Int_val(nb_device));
	d_A = Cl_mem_val(Field(dev_vec, 1));
	size =Int_val(Field(vector, 4));
	type_size = Int_val(Field(Field(customArray, 1),1))*sizeof(char);


	OPENCL_TRY("clGetContextInfo", clGetContextInfo(ctx, CL_CONTEXT_DEVICES, (size_t)sizeof(cl_device_id), &device_id, NULL)) ;

	//OPENCL_CHECK_CALL1(opencl_error, clRetainCommandQueue(queue[Int_val(queue_id)]));
	OPENCL_CHECK_CALL1(opencl_error, clEnqueueReadBuffer(queue[Int_val(queue_id)], d_A, CL_TRUE, 0, size*type_size, h_A, 0, NULL, NULL));
	//OPENCL_CHECK_CALL1(opencl_error, clReleaseCommandQueue(queue[Int_val(queue_id)]));

	OPENCL_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

extern CAMLprim value caml_ba_set_1(value vb, value vind1, value newval);
extern CAMLprim value caml_ba_get_1(value vb, value vind1);



CAMLprim value spoc_set (value vect, value offset, value newval){
	CAMLparam3(vect, offset, newval);
	CAMLlocal1(bigarray);
	bigarray = Field(vect,0);
	caml_ba_set_1(bigarray, offset, newval);
	CAMLreturn(Val_unit);
}


CAMLprim value spoc_get (value vect, value offset){
	CAMLparam2(vect, offset);
	CAMLlocal2(bigarray, ret);
	bigarray = Field(vect,0);
	ret = caml_ba_get_1(bigarray, offset);
	CAMLreturn(ret);
}



CAMLprim value custom_getsizeofbool()
{
	CAMLparam0();
	CAMLreturn(Val_int(sizeof(int)));
}



CAMLprim value custom_boolget (value customArray, value idx)
{
	CAMLparam2(customArray, idx);
	int *b;
	b = ((int*)(Field(customArray, 1)))+(Int_val(idx));
	CAMLreturn(Val_bool(*b));
}

CAMLprim value custom_boolset (value customArray, value idx, value v)
{
	CAMLparam3(customArray, idx, v);
	int* b;
	int i;
	i = Int_val(idx);
	b = ((int*)(Field(customArray, 1)))+i;

	*b = Bool_val(v);
	CAMLreturn(Val_unit);
}



#ifdef __cplusplus
}
#endif

