#ifdef __cplusplus
extern "C" {
#endif


#include "Opencl_dynlink.h"

tclGetDeviceIDs* clGetDeviceIDs;
tclGetPlatformIDs* clGetPlatformIDs;
tclGetPlatformInfo* clGetPlatformInfo;
tclGetDeviceInfo* clGetDeviceInfo;
tclCreateContext* clCreateContext;
tclCreateBuffer* clCreateBuffer;
tclReleaseMemObject* clReleaseMemObject;
tclCreateCommandQueue* clCreateCommandQueue;
tclEnqueueReadBuffer* clEnqueueReadBuffer;
tclEnqueueWriteBuffer* clEnqueueWriteBuffer;
tclGetContextInfo* clGetContextInfo;
tclCreateProgramWithSource* clCreateProgramWithSource;
tclBuildProgram* clBuildProgram;
tclGetProgramBuildInfo* clGetProgramBuildInfo;
tclCreateKernel* clCreateKernel;
tclSetKernelArg* clSetKernelArg;
tclEnqueueNDRangeKernel* clEnqueueNDRangeKernel;
tclFlush* clFlush;
tclFinish* clFinish;
tclRetainCommandQueue* clRetainCommandQueue;
tclReleaseCommandQueue* clReleaseCommandQueue;


#define CL_ERROR_UNKNOWN -12

#ifdef _WIN32
    #include <Windows.h>

    #ifdef UNICODE
    static LPCWSTR __OpenCLLibName = L"opencl.dll";
    #else
    static LPCSTR __OpenCLLibName = "opencl.dll";
    #endif

    typedef HMODULE OPENCLDRIVER;

    static int LOAD_LIBRARY(OPENCLDRIVER *pInstance)
    {
        *pInstance = LoadLibrary(__OpenCLLibName);
        if (*pInstance == NULL)
        {
            printf("LoadLibrary \"%s\" failed!\n", __OpenCLLibName);
            return CL_ERROR_UNKNOWN;
        }
        return CL_SUCCESS;
    }

    #define GET_PROC_EX(name, alias, required)                              \
        alias = (t##name *)GetProcAddress(OpenCLDrvLib, #name);               \
        if (alias == NULL && required) {                                    \
            printf("Failed to find required function \"%s\" in %s\n",       \
                   #name, __OpenCLLibName);                                   \
            return CL_ERROR_UNKNOWN;                                      \
        }

#elif defined(__unix__) || defined(__APPLE__) || defined(__MACOSX)

    #include <dlfcn.h>

    #if defined(__APPLE__) || defined(__MACOSX)
    static char __OpenCLLibName[] = "/System/Library/Frameworks/OpenCL.framework/OpenCL";
    #else
  static char __OpenCLLibName[] = "libOpenCL.so"; //"/home/mathias/ati-stream-sdk-v2.2-lnx64/lib/x86_64/libOpenCL.so";///usr/lib/libOpenCL.so";//libOpenCL.so";
    #endif

    typedef void * OPENCLDRIVER;

    static int  LOAD_LIBRARY(OPENCLDRIVER *pInstance)
    {
        *pInstance = dlopen(__OpenCLLibName, RTLD_NOW);
        if (*pInstance == NULL)
        {
            printf("dlopen \"%s\" failed!\n", __OpenCLLibName);
            return CL_ERROR_UNKNOWN;
        }
        return CL_SUCCESS;
    }

    #define GET_PROC_EX(name, alias, required)                              \
        alias = (t##name *)dlsym(OpenCLDrvLib, #name);                        \
        if (name == NULL && required) {                                    \
            printf("Failed to find required function \"%s\" in %s\n",       \
                    #name, __OpenCLLibName);                                  \
            return CL_ERROR_UNKNOWN;                                      \
        }

#else
#error unsupported platform
#endif

#define CHECKED_CALL(call)              \
    do {                                \
        int result = (call);       \
        if (CL_SUCCESS != result) {   \
            return result;              \
        }                               \
    } while(0)

#define GET_PROC_REQUIRED(name) GET_PROC_EX(name,name,1)
#define GET_PROC_OPTIONAL(name) GET_PROC_EX(name,name,0)
#define GET_PROC(name) GET_PROC_REQUIRED(name)

int noCL = 0;

int CL_API_ENTRY clInit(){
    	OPENCLDRIVER OpenCLDrvLib;
    	//int driverVer = 1000;

    	CHECKED_CALL(LOAD_LIBRARY(&OpenCLDrvLib));


    	GET_PROC(clGetDeviceIDs);
    	GET_PROC(clGetPlatformIDs);
    	GET_PROC(clGetPlatformInfo);
    	GET_PROC(clGetDeviceInfo);
    	GET_PROC(clCreateContext);

    	GET_PROC(clCreateBuffer);
    	GET_PROC(clReleaseMemObject);
    	GET_PROC(clCreateCommandQueue);
    	GET_PROC(clEnqueueReadBuffer);
    	GET_PROC(clEnqueueWriteBuffer);
    	GET_PROC(clGetContextInfo);
    	GET_PROC(clCreateProgramWithSource);
    	GET_PROC(clBuildProgram);
    	GET_PROC(clGetProgramBuildInfo);
    	GET_PROC(clCreateKernel);
    	GET_PROC(clSetKernelArg);
    	GET_PROC(clEnqueueNDRangeKernel);
    	GET_PROC(clFlush);
    	GET_PROC(clFinish);
    	GET_PROC(clRetainCommandQueue);
    	GET_PROC(clReleaseCommandQueue);
    	return CL_SUCCESS;
    	//return CL_ERROR_UNKNOWN;
    }



extern int nbCudaDevices;

value spoc_clInit() {
	CAMLparam0();
	if (CL_SUCCESS != clInit())
	{

		noCL = 1;
 	}
	//if (CUDA_SUCCESS == cudaRuntimeDynload());
	CAMLreturn(Val_unit);
}


#ifdef __cplusplus
}
#endif

