#ifndef _SPOC_H_
#define _SPOC_H_
#include <stdio.h>
#include <stdlib.h>

#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/custom.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <caml/bigarray.h>
#include <caml/signals.h>

#include "Cuda/cuComplex.h"



#include "cuda_drvapi_dynlink_cuda.h"
#include "Opencl_dynlink.h"

/*#ifdef _WIN32
#define THREAD_LOCAL   static  __declspec(thread) 
#else
#define THREAD_LOCAL //static __thread
#endif
*/

int noCL;
int noCuda;
int nbCudaDevices;

//THREAD_LOCAL enum cudaError_enum cuda_error = 0;
//THREAD_LOCAL cl_int opencl_error = 0;
//THREAD_LOCAL int offset = 0;





#define DEBUG 1

#ifdef _WIN32
#define FUNCTION __FUNCTION__
#else
#define FUNCTION __func__
#endif

#ifdef DEBUG
#define PRINT_FUNC \
	printf("IN: %s @ %d\n", FUNCTION,__LINE__ ); fflush(stdout);
#else
#define PRINT_FUNC \
	printf("IN: %s@ %d\n", FUNCTION,__LINE__ ); fflush(stdout);
#endif

#define RAISE_CUDA_ERROR \
	switch (cuda_error){ \
		case  CUDA_ERROR_DEINITIALIZED: \
			raise_constant(*caml_named_value("cuda_error_deinitialized")) ; \
			break; \
		case CUDA_ERROR_NOT_INITIALIZED: \
			raise_constant(*caml_named_value("cuda_error_not_initialized")) ; \
			break; \
		case CUDA_ERROR_INVALID_CONTEXT: \
			raise_constant(*caml_named_value("cuda_error_invalid_context")) ; \
			break; \
		case CUDA_ERROR_INVALID_VALUE: \
			raise_constant(*caml_named_value("cuda_error_invalid_value")) ; \
			break; \
		case CUDA_ERROR_OUT_OF_MEMORY: \
			raise_constant(*caml_named_value("cuda_error_out_of_memory")) ; \
			break; \
		case CUDA_ERROR_INVALID_DEVICE: \
			raise_constant(*caml_named_value("cuda_error_invalid_device")) ; \
			break; \
		case CUDA_ERROR_NOT_FOUND: \
			raise_constant(*caml_named_value("cuda_error_not_found")) ; \
			break; \
		case CUDA_ERROR_FILE_NOT_FOUND: \
			raise_constant(*caml_named_value("cuda_error_file_not_found")) ; \
			break; \
		case CUDA_ERROR_LAUNCH_FAILED: \
			raise_constant(*caml_named_value("cuda_error_launch_failed")) ; \
			break; \
		case CUDA_ERROR_LAUNCH_OUT_OF_RESOURCES: \
			raise_constant(*caml_named_value("cuda_error_launch_out_of_resources")) ; \
			break; \
		case CUDA_ERROR_LAUNCH_TIMEOUT: \
			raise_constant(*caml_named_value("cuda_error_launch_timeout")) ; \
			break; \
		case CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING: \
			raise_constant(*caml_named_value("cuda_error_launch_incompatible_texturing")) ; \
			break; \
		case CUDA_ERROR_UNKNOWN: \
		default: \
			raise_constant(*caml_named_value("cuda_error_unknown")) ; \
			break;\
	}\


#define CUDA_CHECK_CALL(func) \
		cuda_error = func; \
		if (CUDA_SUCCESS != cuda_error ) { \
			printf("IN: %s@ %d\n", FUNCTION,__LINE__ ); fflush(stdout); \
			RAISE_CUDA_ERROR\
		}

#define ALIGN_UP(offset, alignment) \
       (offset) = ((offset) + (alignment) - 1) & ~((alignment) - 1)

#define UNIMPLENTED \
		printf("ERROR ************* %s unimplented\n", FUNCTION); \
		fflush(stdout);




#define OPENCL_TRY(name,fun)                                    \
  {                                                             \
    cl_int err = fun;                                              \
    if (err != CL_SUCCESS) {                                       \
      fprintf(stderr,"ERROR %d calling %s().\n", err,name);             \
      return -1;                                                        \
    }                                                                   \
  }


#define CL_FILE_NOT_FOUND -1

#define RAISE_OPENCL_ERROR \
	switch (opencl_error){ \
		case CL_INVALID_CONTEXT: raise_constant(*caml_named_value("opencl_invalid_context")) ; \
			break; \
		case CL_INVALID_DEVICE: raise_constant(*caml_named_value("opencl_invalid_device")) ; \
			break; \
		case CL_INVALID_VALUE: raise_constant(*caml_named_value("opencl_invalid_value")) ; \
			break; \
		case CL_INVALID_QUEUE_PROPERTIES: raise_constant(*caml_named_value("opencl_invalid_queue_properties")) ; \
			break; \
		case CL_OUT_OF_RESOURCES: raise_constant(*caml_named_value("opencl_out_of_resources")) ; \
			break; \
		case CL_MEM_OBJECT_ALLOCATION_FAILURE:	raise_constant(*caml_named_value("opencl_mem_object_allocation_failure")) ; \
			break; \
		case CL_OUT_OF_HOST_MEMORY: raise_constant(*caml_named_value("opencl_out_of_host_memory")) ; \
			break; \
		case CL_INVALID_PROGRAM: raise_constant(*caml_named_value("opencl_invalid_program")) ; \
			break; \
		case CL_INVALID_BINARY: raise_constant(*caml_named_value("opencl_invalid_binary")) ; \
			break; \
		case CL_INVALID_BUILD_OPTIONS: raise_constant(*caml_named_value("opencl_invalid_build_options")) ; \
			break; \
		case CL_INVALID_OPERATION: raise_constant(*caml_named_value("opencl_invalid_operation")) ; \
			break; \
		case CL_COMPILER_NOT_AVAILABLE: raise_constant(*caml_named_value("opencl_compiler_not_available")) ; \
			break; \
		case CL_BUILD_PROGRAM_FAILURE: raise_constant(*caml_named_value("opencl_build_program_failure")) ; \
			break; \
		case CL_INVALID_KERNEL: raise_constant(*caml_named_value("opencl_invalid_kernel")) ; \
			break; \
		case CL_INVALID_ARG_INDEX: raise_constant(*caml_named_value("opencl_invalid_arg_index")) ; \
			break; \
		case CL_INVALID_ARG_VALUE: raise_constant(*caml_named_value("opencl_invalid_arg_value")) ; \
			break; \
		case CL_INVALID_MEM_OBJECT: raise_constant(*caml_named_value("opencl_invalid_mem_object")) ; \
			break; \
		case CL_INVALID_SAMPLER: raise_constant(*caml_named_value("opencl_invalid_sampler")) ; \
			break; \
		case CL_INVALID_ARG_SIZE: raise_constant(*caml_named_value("opencl_invalid_arg_size")) ; \
			break; \
		case CL_FILE_NOT_FOUND: raise_constant(*caml_named_value("opencl_file_not_found")) ; \
			break; \
		case CL_INVALID_COMMAND_QUEUE: raise_constant(*caml_named_value("opencl_invalid_command_queue")) ; \
			break; \
		default: raise_constant(*caml_named_value("opencl_error_unknown")) ; \
			break; \
}



#define OPENCL_CHECK_CALL1(val, fun) \
		val = fun; \
		if (CL_SUCCESS != opencl_error) { \
			printf("IN: %s@ %d\n", FUNCTION,__LINE__ ); fflush(stdout); \
			RAISE_OPENCL_ERROR \
		}



#define CUDA_GET_CONTEXT \
	CUcontext ctx; \
	enum cudaError_enum cuda_error = 0; \
	ctx = (CUcontext)Field(gi, 9); \
	CUDA_CHECK_CALL(cuCtxSetCurrent(ctx)); \
	CUDA_CHECK_CALL(cuCtxAttach(&ctx, 0));

#define CUDA_RESTORE_CONTEXT \
	CUDA_CHECK_CALL(cuCtxDetach(ctx)); \
	Store_field(gi,9, (value)ctx);

typedef struct spoc_cl_context {
		cl_context ctx;
		cl_command_queue queue[2];
}spoc_cl_context;


#define OPENCL_GET_CONTEXT \
	cl_context ctx; \
	cl_command_queue queue[2]; \
	spoc_cl_context *spoc_ctx; \
	cl_int opencl_error = 0; \
	spoc_ctx = (spoc_cl_context*)Field(gi, 9); \
	ctx = spoc_ctx->ctx; \
	queue[0] = spoc_ctx->queue[0];\
	queue[1] = spoc_ctx->queue[1];\
	caml_enter_blocking_section();



#define OPENCL_RESTORE_CONTEXT \
	caml_leave_blocking_section(); \
	spoc_ctx->queue[0] = queue[0]; \
	spoc_ctx->queue[1] = queue[1]; \
	Store_field(gi,9, (value)spoc_ctx);


int ae_load_file_to_memory(const char *filename, char **result);


#define GET_TYPE_SIZE \
	tag = Bigarray_val(bigArray)->flags & BIGARRAY_KIND_MASK; \
	switch (tag) { \
	case BIGARRAY_FLOAT32 : \
		type_size = sizeof(float); \
		break; \
	case BIGARRAY_FLOAT64 : \
		type_size = sizeof(double); \
		break; \
	case BIGARRAY_INT32 : \
		type_size = sizeof(int); \
		break; \
	case BIGARRAY_INT64 : \
		type_size = sizeof(int)*2; \
		break; \
	case BIGARRAY_COMPLEX32: \
		type_size = sizeof(float)*2;\
		break; \
	}

#define Val_cl_mem(x) (value)((cl_mem)x)
#define Cl_mem_val(x) (cl_mem)((value)x)

#define Val_CUdeviceptr(x) (value)((CUdeviceptr)x)
#define CUdeviceptr_val(x) (CUdeviceptr)((value)x)

cuComplex complex_val (value c);
cuDoubleComplex doubleComplex_val (value c);
value copy_two_doubles(double d0, double d1);
value copy_complex(cuComplex c);
value copy_doubleComplex(cuDoubleComplex c);

#define Complex_val(c) complex_val(c)


#endif
