#ifndef _SPOC_CLAMDBLAS_H_
#define _SPOC_CL_AMD_BLAS_H_

#include <stdio.h>
#include <stdlib.h>

#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/custom.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <caml/bigarray.h>
#include <caml/signals.h>


#include <clAmdBlas.h>

typedef struct spoc_cl_context {
		cl_context ctx;
		cl_command_queue queue[2];
}spoc_cl_context;

#define DEBUG 1

#ifdef _WIN32
#define FUNCTION __FUNCTION__
#else
#define FUNCTION __func__
#endif

#ifdef DEBUG
#define PRINT_FUNC \
	printf("IN: %s @ %d\n", FUNCTION,__LINE__ ); fflush(stdout);
#else
#define PRINT_FUNC \
	printf("IN: %s@ %d\n", FUNCTION,__LINE__ ); fflush(stdout);
#endif

#define OPENCL_TRY(name,fun)                                    \
  {                                                             \
    cl_int err = fun;                                              \
    if (err != CL_SUCCESS) {                                       \
      fprintf(stderr,"ERROR %d calling %s().\n", err,name);             \
      return -1;                                                        \
    }                                                                   \
  }


#define CL_FILE_NOT_FOUND -1

#define RAISE_OPENCL_ERROR \
	switch (opencl_error){ \
		case CL_INVALID_CONTEXT: raise_constant(*caml_named_value("opencl_invalid_context")) ; \
			break; \
		case CL_INVALID_DEVICE: raise_constant(*caml_named_value("opencl_invalid_device")) ; \
			break; \
		case CL_INVALID_VALUE: raise_constant(*caml_named_value("opencl_invalid_value")) ; \
			break; \
		case CL_INVALID_QUEUE_PROPERTIES: raise_constant(*caml_named_value("opencl_invalid_queue_properties")) ; \
			break; \
		case CL_OUT_OF_RESOURCES: raise_constant(*caml_named_value("opencl_out_of_resources")) ; \
			break; \
		case CL_MEM_OBJECT_ALLOCATION_FAILURE:	raise_constant(*caml_named_value("opencl_mem_object_allocation_failure")) ; \
			break; \
		case CL_OUT_OF_HOST_MEMORY: raise_constant(*caml_named_value("opencl_out_of_host_memory")) ; \
			break; \
		case CL_INVALID_PROGRAM: raise_constant(*caml_named_value("opencl_invalid_program")) ; \
			break; \
		case CL_INVALID_BINARY: raise_constant(*caml_named_value("opencl_invalid_binary")) ; \
			break; \
		case CL_INVALID_BUILD_OPTIONS: raise_constant(*caml_named_value("opencl_invalid_build_options")) ; \
			break; \
		case CL_INVALID_OPERATION: raise_constant(*caml_named_value("opencl_invalid_operation")) ; \
			break; \
		case CL_COMPILER_NOT_AVAILABLE: raise_constant(*caml_named_value("opencl_compiler_not_available")) ; \
			break; \
		case CL_BUILD_PROGRAM_FAILURE: raise_constant(*caml_named_value("opencl_build_program_failure")) ; \
			break; \
		case CL_INVALID_KERNEL: raise_constant(*caml_named_value("opencl_invalid_kernel")) ; \
			break; \
		case CL_INVALID_ARG_INDEX: raise_constant(*caml_named_value("opencl_invalid_arg_index")) ; \
			break; \
		case CL_INVALID_ARG_VALUE: raise_constant(*caml_named_value("opencl_invalid_arg_value")) ; \
			break; \
		case CL_INVALID_MEM_OBJECT: raise_constant(*caml_named_value("opencl_invalid_mem_object")) ; \
			break; \
		case CL_INVALID_SAMPLER: raise_constant(*caml_named_value("opencl_invalid_sampler")) ; \
			break; \
		case CL_INVALID_ARG_SIZE: raise_constant(*caml_named_value("opencl_invalid_arg_size")) ; \
			break; \
		case CL_FILE_NOT_FOUND: raise_constant(*caml_named_value("opencl_file_not_found")) ; \
			break; \
		case CL_INVALID_COMMAND_QUEUE: raise_constant(*caml_named_value("opencl_invalid_command_queue")) ; \
			break; \
		default: raise_constant(*caml_named_value("opencl_error_unknown")) ; \
			break; \
}



#define OPENCL_CHECK_CALL1(val, fun) \
		val = fun; \
		if (CL_SUCCESS != opencl_error) { \
			printf("IN: %s@ %d\n", FUNCTION,__LINE__ ); fflush(stdout); \
			RAISE_OPENCL_ERROR \
		}


#define CLAMDBLAS_GET_CONTEXT \
	cl_context ctx; \
	cl_command_queue queue[2]; \
	spoc_cl_context *spoc_ctx; \
	spoc_ctx = (spoc_cl_context*)Field(gi, 9); \
	ctx = spoc_ctx->ctx; \
	queue[0] = spoc_ctx->queue[0];\
	queue[1] = spoc_ctx->queue[1];\
	caml_enter_blocking_section();



#define CLAMDBLAS_RESTORE_CONTEXT \
	caml_leave_blocking_section(); \
	spoc_ctx->queue[0] = queue[0]; \
	spoc_ctx->queue[1] = queue[1]; \
	Store_field(gi,9, (value)spoc_ctx);

#define Val_cl_mem(x) (value)((cl_mem)x)
#define Cl_mem_val(x) (cl_mem)((value)x)


#endif
