#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/custom.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <caml/bigarray.h>
#include <math.h>
#include "cuda_drvapi_dynlink_cuda.h"
#include "Spoc.h"

value spoc_cuInit() {

	CAMLparam0();
	//CUdevice dev;
	if ((CUDA_SUCCESS != cuInit(0, 4000)))
/*		&& (CUDA_SUCCESS != cuInit(0, 3020))
		&& (CUDA_SUCCESS != cuInit(0, 3010))
		&& (CUDA_SUCCESS != cuInit(0, 3000))
		&& (CUDA_SUCCESS != cuInit(0, 2030))
		&& (CUDA_SUCCESS != cuInit(0, 2010))
		&& (CUDA_SUCCESS != cuInit(0, 2000))
		&& (CUDA_SUCCESS != cuInit(0, 1000)))*/
	{
		noCuda = 1;
		CAMLreturn(Val_unit);
	}
	else noCuda = 0;
	//CUDA_CHECK_CALL(cuDeviceGet(&dev, 0));



	CAMLreturn(Val_unit);
}


value spoc_getCudaDevicesCount()
{
	CAMLparam0();
	int nb_devices;
	if (noCuda) CAMLreturn (Val_int(0));
    cuDeviceGetCount (&nb_devices);
    nbCudaDevices = nb_devices;
    CAMLreturn (Val_int(nb_devices));

}


value spoc_getCudaDevice(value i)
{
	CAMLparam1(i);
	CAMLlocal3(general_info, cuda_info, specific_info);
	CAMLlocal3(device,  maxT, maxG);
	int nb_devices;
	CUdevprop dev_infos;
	CUdevice dev;
	CUcontext ctx;
	char infoStr[1024];
	int infoInt;
	size_t infoUInt;
	int major, minor;
	enum cudaError_enum cuda_error; 


	cuDeviceGetCount (&nb_devices);

	if ((Int_val(i)) > nb_devices)
		raise_constant(*caml_named_value("no_cuda_device")) ;


	cuDeviceGet(&dev, Int_val(i));
	cuDeviceGetProperties(&dev_infos, dev);

	general_info = caml_alloc (10, 0);
	cuDeviceGetName(infoStr, sizeof(infoStr), dev);

	Store_field(general_info,0, copy_string(infoStr));//
	cuDeviceTotalMem(&infoUInt, dev);

	Store_field(general_info,1, Val_int(infoUInt));//
	Store_field(general_info,2, Val_int(dev_infos.sharedMemPerBlock));//
	Store_field(general_info,3, Val_int(dev_infos.clockRate));//
	Store_field(general_info,4, Val_int(dev_infos.totalConstantMemory));//
	cuDeviceGetAttribute(&infoInt, CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT, dev);
	Store_field(general_info,5, Val_int(infoInt));//
	cuDeviceGetAttribute(&infoInt, CU_DEVICE_ATTRIBUTE_ECC_ENABLED, dev);
	Store_field(general_info,6, Val_bool(infoInt));//
	Store_field(general_info,7, Val_int(0)); //0 -> Cuda
	Store_field(general_info,8, i);
	CUDA_CHECK_CALL(cuCtxCreate	(&ctx, CU_CTX_BLOCKING_SYNC, dev));
	//CUDA_CHECK_CALL(cuCtxGetCurrent(&ctx));
	Store_field(general_info,9, (value)ctx);


	cuda_info = caml_alloc(1, 0); //0 -> Cuda
	specific_info = caml_alloc(18, 0);

	cuDeviceComputeCapability(&major, &minor, dev);
	Store_field(specific_info,0, Val_int(major));//
	Store_field(specific_info,1, Val_int(minor));//
	Store_field(specific_info,2, Val_int(dev_infos.regsPerBlock));//
	Store_field(specific_info,3, Val_int(dev_infos.SIMDWidth));//
	Store_field(specific_info,4, Val_int(dev_infos.memPitch));//
	Store_field(specific_info,5, Val_int(dev_infos.maxThreadsPerBlock));//

	maxT = caml_alloc(3, 0);
	Store_field(maxT,0, Val_int(dev_infos.maxThreadsDim[0]));//
	Store_field(maxT,1, Val_int(dev_infos.maxThreadsDim[1]));//
	Store_field(maxT,2, Val_int(dev_infos.maxThreadsDim[2]));//
	Store_field(specific_info,6, maxT);

	maxG = caml_alloc(3, 0);
	Store_field(maxG,0, Val_int(dev_infos.maxGridSize[0]));//
	Store_field(maxG,1, Val_int(dev_infos.maxGridSize[1]));//
	Store_field(maxG,2, Val_int(dev_infos.maxGridSize[2]));//
	Store_field(specific_info,7, maxG);

	Store_field(specific_info,8, Val_int(dev_infos.textureAlign));//
	cuDeviceGetAttribute(&infoInt, CU_DEVICE_ATTRIBUTE_GPU_OVERLAP, dev);
	Store_field(specific_info,9, Val_bool(infoInt));//
	cuDeviceGetAttribute(&infoInt, CU_DEVICE_ATTRIBUTE_KERNEL_EXEC_TIMEOUT, dev);
	Store_field(specific_info,10, Val_bool(infoInt));//
	cuDeviceGetAttribute(&infoInt, CU_DEVICE_ATTRIBUTE_INTEGRATED, dev);
	Store_field(specific_info,11, Val_bool(infoInt));//
	cuDeviceGetAttribute(&infoInt, CU_DEVICE_ATTRIBUTE_CAN_MAP_HOST_MEMORY, dev);
	Store_field(specific_info,12, Val_bool(infoInt));//
	cuDeviceGetAttribute(&infoInt, CU_DEVICE_ATTRIBUTE_COMPUTE_MODE, dev);
	Store_field(specific_info,13, Val_int(infoInt));//
	cuDeviceGetAttribute(&infoInt, CU_DEVICE_ATTRIBUTE_CONCURRENT_KERNELS, dev);
	Store_field(specific_info,14, Val_bool(infoInt));//
	cuDeviceGetAttribute(&infoInt, CU_DEVICE_ATTRIBUTE_PCI_BUS_ID, dev);
	Store_field(specific_info,15, Val_int(infoInt));
	cuDeviceGetAttribute(&infoInt, CU_DEVICE_ATTRIBUTE_PCI_DEVICE_ID, dev);
	Store_field(specific_info,16, Val_int(infoInt));


	Store_field(cuda_info, 0, specific_info);
	device = caml_alloc(2, 0);
	Store_field(device, 0, general_info);
	Store_field(device, 1, cuda_info);

	CAMLreturn(device);
}

#ifdef __cplusplus
}
#endif
