#include "Spoc_cublas.h"

//cublas library

CAMLprim value spoc_cublasInit(){
	cublasStatus cublas_error = CUBLAS_STATUS_SUCCESS;
	CUBLAS_CHECK_CALL(cublasInit());
	return Val_unit;
}

CAMLprim value spoc_cublasShutdown(){
	cublasStatus cublas_error= CUBLAS_STATUS_SUCCESS;
	CUBLAS_CHECK_CALL(cublasShutdown());
	return Val_unit;
}

CAMLprim value spoc_cublasGetError(){
	cublasStatus cublas_error = CUBLAS_STATUS_SUCCESS;
	CUBLAS_CHECK_CALL(cublasGetError());
	return Val_unit;
}

#define GET_HOST_VEC(vec, h_vec) \
	bigArray = Field(vec, 1); \
	h_vec = Data_bigarray_val(bigArray);

#define GET_VEC(vec, d_vec) \
	dev_vec_array = Field(vec, 2); \
	dev_vec =Field(dev_vec_array, id); \
	d_vec = CUdeviceptr_val(Field(dev_vec, 1));

#define CUBLAS_FUN(fun, arg, arg2) \
		CAMLprim value fun##_n arg {\
			return fun arg2 ;\
		}\
		CAMLprim value fun##_b arg {\
			return fun arg2 ;\
		}
/******************************* BLAS1 ***************************************/

/* Single-Precision BLAS1 functions */

CAMLprim value spoc_cublasIsamax(value n, value x, value incx, value dev){
	CAMLparam4(n,x,incx, dev);
	CAMLlocal3(dev_vec_array, dev_vec, gi);
	int res;
	CUdeviceptr d_A;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	res = cublasIsamax(Int_val(n), (float*)d_A, Int_val(incx));
	CUBLAS_CHECK_CALL(cublasGetError());
	CUDA_RESTORE_CONTEXT;

	CAMLreturn(Val_int(res));
}

CAMLprim value spoc_cublasIsamin(value n, value x, value incx, value dev){
	CAMLparam4(n,x,incx, dev);
	CAMLlocal3(dev_vec_array, dev_vec, gi);
	int res;
	CUdeviceptr d_A;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	res = cublasIsamin(Int_val(n), (float*)d_A, Int_val(incx));
	CUBLAS_CHECK_CALL(cublasGetError());
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_int(res));
}

CAMLprim value spoc_cublasSasum (value n, value x, value incx, value dev){
	CAMLparam4(n,x,incx, dev);
	CAMLlocal4(dev_vec_array, dev_vec, res, gi);
	CUdeviceptr d_A;
	float result;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	result = cublasSasum(Int_val(n), (float*)d_A, Int_val(incx));
	CUBLAS_CHECK_CALL(cublasGetError());
	res = caml_copy_double((double)result);
	CUDA_RESTORE_CONTEXT;
	CAMLreturn((res));
}

CAMLprim value spoc_cublasSaxpy (value n, value alpha, value x, value incx, value y, value incy, value dev){
	CAMLparam5(n,alpha, x,incx, y);
	CAMLxparam2(incy, dev);
	CAMLlocal3(dev_vec_array, dev_vec, gi);
	CUdeviceptr d_A;
	CUdeviceptr d_B;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	GET_VEC(y, d_B);
	cublasSaxpy(Int_val(n), (float)(Double_val(alpha)), (float*)d_A, Int_val(incx), (float*)d_B, Int_val(incy));
	CUBLAS_CHECK_CALL(cublasGetError());
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cublasScopy (value n, value x, value incx, value y, value incy, value dev){
	CAMLparam5(n,x,incx, y, incy);
	CAMLxparam1(dev);
	CAMLlocal3(dev_vec_array, dev_vec, gi);
	CUdeviceptr d_A;
	CUdeviceptr d_B;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	GET_VEC(y, d_B);
	cublasScopy(Int_val(n), (float*)d_A, Int_val(incx), (float*)d_B, Int_val(incy));
	CUBLAS_CHECK_CALL(cublasGetError());
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cublasSdot (value n, value x, value incx, value y, value incy, value dev){
	CAMLparam5(n,x,incx, y, incy);
	CAMLxparam1(dev);
	CAMLlocal4(dev_vec_array, dev_vec, res, gi);
	float result;
	CUdeviceptr d_A;
	CUdeviceptr d_B;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	GET_VEC(y, d_B);
	result = cublasSdot(Int_val(n), (float*)d_A, Int_val(incx), (float*)d_B, Int_val(incy));
	CUBLAS_CHECK_CALL(cublasGetError());
	res = caml_copy_double((double)result);
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(res);
}

CAMLprim value spoc_cublasSnrm2 (value n, value x, value incx, value dev){
	CAMLparam4(n,x,incx, dev);
	CAMLlocal4(dev_vec_array, dev_vec, res, gi);
	CUdeviceptr d_A;
	float result;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	result = cublasSnrm2(Int_val(n), (float*)d_A, Int_val(incx));
	CUBLAS_CHECK_CALL(cublasGetError());
	res = caml_copy_double((double)result);
	CUDA_RESTORE_CONTEXT;
	CAMLreturn((res));
}

CAMLprim value spoc_cublasSrot (value n, value x, value incx, value y, value incy, value sc, value ss, value dev){
	CAMLparam5(n,x,incx, y, incy);
	CAMLxparam3(sc, ss, dev);
	CAMLlocal4(dev_vec_array, dev_vec, res, gi);
	CUdeviceptr d_A;
	CUdeviceptr d_B;
	float result;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	GET_VEC(y, d_B);
	cublasSrot(Int_val(n), (float*)d_A, Int_val(incx), (float*)d_B, Int_val(incy), (float)(Double_val(sc)), (float)(Double_val(ss)));
	CUBLAS_CHECK_CALL(cublasGetError());
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cublasSrotg (value host_sa, value host_sb, value host_sc, value host_ss){
	CAMLparam4(host_sa, host_sb, host_sc, host_ss);
	CAMLlocal2(bigArray, gi);
	int id;
	enum cudaError_enum cuda_error = 0;
	cublasStatus cublas_error = CUBLAS_STATUS_SUCCESS;
	float* h_A;
	float* h_B;
	float* h_C;
	float* h_D;
	float result;
	GET_HOST_VEC(host_sa, h_A);
	GET_HOST_VEC(host_sb, h_B);
	GET_HOST_VEC(host_sc, h_C);
	GET_HOST_VEC(host_ss, h_D);
	cublasSrotg(h_A, h_B, h_C, h_D);
	CUBLAS_CHECK_CALL(cublasGetError());
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cublasSrotm (value n, value x, value incx, value y, value incy, value sparam, value dev){
	CAMLparam5(n,x,incx, y, incy);
	CAMLxparam2(sparam, dev);
	CAMLlocal4(dev_vec_array, dev_vec, res, gi);
	CUdeviceptr d_A;
	CUdeviceptr d_B;
	CUdeviceptr d_C;
	float result;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	GET_VEC(y, d_B);
	GET_VEC(sparam, d_C);
	cublasSrotm(Int_val(n), (float*)d_A, Int_val(incx), (float*)d_B, Int_val(incy), (float*)sparam);
	CUBLAS_CHECK_CALL(cublasGetError());
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cublasSrotmg (value host_psd1, value host_psd2, value host_psx1, value host_psy1, value host_sparam){
	CAMLparam5(host_psd1, host_psd2, host_psx1, host_psy1, host_sparam);
	CAMLlocal2(bigArray, gi);
	int id;
	enum cudaError_enum cuda_error = 0;
	cublasStatus cublas_error = CUBLAS_STATUS_SUCCESS;
	float* h_A;
	float* h_B;
	float* h_C;
	float* h_D;
	float* h_E;
	GET_HOST_VEC(host_psd1, h_A);
	GET_HOST_VEC(host_psd2, h_B);
	GET_HOST_VEC(host_psx1, h_C);
	GET_HOST_VEC(host_psy1, h_D);
	GET_HOST_VEC(host_sparam, h_E);
	cublasSrotmg(h_A, h_B, h_C, h_D, h_E);
	CUBLAS_CHECK_CALL(cublasGetError());
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cublasSscal (value n, value alpha, value x, value incx, value dev){
	CAMLparam5(n, alpha, x,incx, dev);
	CAMLlocal3(dev_vec_array, dev_vec, gi);
	CUdeviceptr d_A;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	cublasSscal(Int_val(n), (float)(Double_val(alpha)), (float*)d_A, Int_val(incx));
	CUBLAS_CHECK_CALL(cublasGetError());
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CAMLprim value spoc_cublasSswap (value n, value x, value incx, value y, value incy, value dev){
	CAMLparam5(n, x,incx, y, incy);
	CAMLxparam1(dev);
	CAMLlocal3(dev_vec_array, dev_vec, gi);
	CUdeviceptr d_A;
	CUdeviceptr d_B;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	GET_VEC(y, d_B);
	cublasSswap(Int_val(n), (float*)d_A, Int_val(incx), (float*)d_B, Int_val(incy));
	CUBLAS_CHECK_CALL(cublasGetError());
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}

CUBLAS_FUN(spoc_cublasIsamax, (value n, value x, value incx, value dev), (n, x, incx, dev));
CUBLAS_FUN(spoc_cublasIsamin, (value n, value x, value incx, value dev), (n, x, incx, dev));
CUBLAS_FUN(spoc_cublasSasum, (value n, value x, value incx, value dev), (n, x, incx, dev));
CUBLAS_FUN(spoc_cublasSaxpy, (value n, value alpha, value x, value incx, value y, value incy, value dev), (n, alpha, x, incx, y, incy, dev));
CUBLAS_FUN(spoc_cublasScopy, (value n, value x, value incx, value y, value incy, value dev), (n, x, incx, y, incy, dev));
CUBLAS_FUN(spoc_cublasSdot, (value n, value x, value incx, value y, value incy, value dev), (n, x, incx, y, incy, dev));
CUBLAS_FUN(spoc_cublasSnrm2, (value n, value x, value incx, value dev), (n, x, incx, dev));
CUBLAS_FUN(spoc_cublasSrot, (value n, value x, value incx, value y, value incy, value sc, value ss, value dev), (n, x, incx, y, incy, sc, ss, dev));
CUBLAS_FUN(spoc_cublasSrotg, (value host_sa, value host_sb, value host_sc, value host_ss), (host_sa, host_sb, host_sc, host_ss));
CUBLAS_FUN(spoc_cublasSrotm, (value n, value x, value incx, value y, value incy, value sparam, value dev), (n, x, incx, y, incy, sparam, dev));
CUBLAS_FUN(spoc_cublasSrotmg, (value host_psd1, value host_psd2, value host_psx1, value host_psy1, value host_sparam), (host_psd1, host_psd2, host_psx1, host_psy1, host_sparam));
CUBLAS_FUN(spoc_cublasSscal, (value n, value alpha, value x, value incx, value dev), (n, alpha, x, incx, dev));
CUBLAS_FUN(spoc_cublasSswap, (value n, value x, value incx, value y, value incy, value dev), (n, x, incx, y, incy, dev));

/* Single-Precision Complex BLAS1 functions */


CAMLprim value spoc_cublasCaxpy (value n, value alpha, value x, value incx, value y, value incy, value dev){
	CAMLparam5(n,alpha, x,incx, y);
	CAMLxparam2(incy, dev);
	CAMLlocal3(dev_vec_array, dev_vec, gi);
	CUdeviceptr d_A;
	CUdeviceptr d_B;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	GET_VEC(y, d_B);
	cublasCaxpy(Int_val(n), Complex_val(alpha), (cuComplex*)d_A, Int_val(incx), (cuComplex*)d_B, Int_val(incy));
	CUBLAS_CHECK_CALL(cublasGetError());
	CUDA_RESTORE_CONTEXT;
	CAMLreturn(Val_unit);
}


CAMLprim value spoc_cublasScasum (value n, value x, value incx, value dev){
	CAMLparam4(n,x,incx, dev);
	CAMLlocal4(dev_vec_array, dev_vec, res, gi);
	CUdeviceptr d_A;
	float result;
	CUBLAS_GET_CONTEXT;
	GET_VEC(x, d_A);
	result = cublasScasum(Int_val(n), (cuComplex*)d_A, Int_val(incx));
	CUBLAS_CHECK_CALL(cublasGetError());
	res = caml_copy_double((double)result);
	CUDA_RESTORE_CONTEXT;
	CAMLreturn((res));
}

CUBLAS_FUN(spoc_cublasCaxpy, (value n, value alpha, value x, value incx, value y, value incy, value dev), (n, alpha, x, incx, y, incy, dev));
CUBLAS_FUN(spoc_cublasScasum, (value n, value x, value incx, value dev), (n, x, incx, dev));


/******************************* BLAS3 ***************************************/

/* Single-Precision BLAS3 functions */



CAMLprim value spoc_cublasSgemm(value transa, value transb,
		value m, value n, value k,
		value alpha, value a, value lda,
		value b, value ldb, value beta, value c, value ldc, value dev){
	CAMLparam5(transa, transb, m, n, k);
	CAMLxparam5(alpha, a, lda, b, ldb);
	CAMLxparam4(beta, c, ldc, dev);
	CAMLlocal3(dev_vec_array, dev_vec, gi);
	CUdeviceptr d_A;
	CUdeviceptr d_B;
	CUdeviceptr d_C;
	CUBLAS_GET_CONTEXT;

	GET_VEC(a, d_A);
	GET_VEC(b, d_B);
	GET_VEC(c, d_C);

	cublasSgemm (Int_val(transa), Int_val(transb), Int_val(m), Int_val(n),
	Int_val(k), (float)Double_val(alpha), (float*) d_A, Int_val(lda),
	(float*) d_B, Int_val(ldb), (float) Double_val(beta),
	(float *)d_C, Int_val(ldc));

	CUDA_RESTORE_CONTEXT;

	CAMLreturn(Val_unit);
}

CUBLAS_FUN(spoc_cublasSgemm, (value transa, value transb, value m, value n, value k, value alpha, value a, value lda, value b, value ldb, value beta, value c, value ldc, value dev),(transa, transb, m, n, k, alpha, a, lda, b, ldb, beta, c, ldc, dev) );
