#ifndef _SPOC_CUBLAS_H_
#define _SPOC_CUBLAS_H_

#include <stdio.h>
#include <stdlib.h>

#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/custom.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <caml/bigarray.h>
#include <caml/signals.h>

#include <cublas.h>

#include "Spoc.h"

#define DEBUG 1

#ifdef _WIN32
#define FUNCTION __FUNCTION__
#else
#define FUNCTION __func__
#endif

#ifdef DEBUG
#define PRINT_FUNC \
	printf("IN: %s @ %d\n", FUNCTION,__LINE__ ); fflush(stdout);
#else
#define PRINT_FUNC \
	printf("IN: %s@ %d\n", FUNCTION,__LINE__ ); fflush(stdout);
#endif


#define RAISE_CUBLAS_ERROR \
	switch (cublas_error){ \
		case  CUBLAS_STATUS_NOT_INITIALIZED: \
			raise_constant(*caml_named_value("cublas_status_not_initialised")) ; \
			break; \
		case  CUBLAS_STATUS_ALLOC_FAILED: \
			raise_constant(*caml_named_value("cublas_status_alloc_failed")) ; \
			break; \
		case  CUBLAS_STATUS_INVALID_VALUE: \
			raise_constant(*caml_named_value("cublas_status_invalid_value")) ; \
			break; \
		case  CUBLAS_STATUS_ARCH_MISMATCH: \
			raise_constant(*caml_named_value("cublas_status_arch_mismatch")) ; \
			break; \
		case  CUBLAS_STATUS_MAPPING_ERROR: \
			raise_constant(*caml_named_value("cublas_status_mapping_error")) ; \
			break; \
		case  CUBLAS_STATUS_EXECUTION_FAILED: \
			raise_constant(*caml_named_value("cublas_status_execution_failed")) ; \
			break; \
		case  CUBLAS_STATUS_INTERNAL_ERROR: \
			raise_constant(*caml_named_value("cublas_status_internal_error")) ; \
			break; \
		default: \
			raise_constant(*caml_named_value("cublas_error_unknown")) ; \
			break;\
	}\

#define CUBLAS_CHECK_CALL(func) \
		cublas_error = func; \
		if (CUBLAS_STATUS_SUCCESS != cublas_error ) { \
			printf("IN: %s@ %d\n", FUNCTION,__LINE__ ); fflush(stdout); \
			RAISE_CUBLAS_ERROR\
		}

#define CUBLAS_GET_CONTEXT \
	CUcontext ctx; \
	int id; \
	enum cudaError_enum cuda_error = 0; \
	cublasStatus cublas_error = CUBLAS_STATUS_SUCCESS; \
	gi = Field(dev, 0); \
	id = Int_val(Field(gi, 8)); \
	ctx = (CUcontext)Field(gi, 9); \
	CUDA_CHECK_CALL(cuCtxSetCurrent(ctx)); \
	CUDA_CHECK_CALL(cuCtxAttach(&ctx, 0));

#define CUBLAS_RESTORE_CONTEXT \
	CUDA_CHECK_CALL(cuCtxDetach(ctx)); \
	Store_field(gi,9, (value)ctx);


#endif
