open Vector
  
open Mem
  
open Devices
  
open Kernel
  
external sizeofbool : unit -> int = "custom_getsizeofbool"
  
external boolget : customarray -> int -> bool = "custom_boolget"
  
external boolset : customarray -> int -> bool -> unit = "custom_boolset"
  
	
let print s = (Printf.printf "%s\n" (string_of_int s); Pervasives.flush stdout)
	
	
let iter f vect =
  ((match vect.dev with
    | No_dev -> ()
    | Dev d ->
        (Devices.flush d ~queue_id: 0 ();
         Devices.flush d ~queue_id: 1 ();
         to_cpu vect ();
         Devices.flush d ~queue_id: 0 ()));
   for idx = 0 to (length vect) - 1 do f (unsafe_get vect idx) done)
  
let iteri f vect =
  ((match vect.dev with
    | No_dev -> ()
    | Dev d ->
        (Devices.flush d ~queue_id: 0 ();
         Devices.flush d ~queue_id: 1 ();
         to_cpu vect ();
         Devices.flush d ~queue_id: 0 ()));
   for idx = 0 to (length vect) - 1 do f (unsafe_get vect idx) idx done)
  
let map f kind vect =
  ((match vect.dev with
    | No_dev -> ()
    | Dev d ->
        (Devices.flush d ~queue_id: 0 ();
         Devices.flush d ~queue_id: 1 ();
         to_cpu vect ();
         Devices.flush d ~queue_id: 0 ()));
   let newvect = Vector.create kind vect.Vector.length
   in
     (for i = 0 to vect.Vector.length - 1 do
        Mem.unsafe_set newvect i (f (Mem.unsafe_get vect i))
      done;
      newvect))
  
let trueCustom =
  { elt = true; size = sizeofbool (); get = boolget; set = boolset; }
  
let falseCustom =
  { elt = false; size = sizeofbool (); get = boolget; set = boolset; }
  
let fold_left f seed vect =
  ((match vect.dev with
    | No_dev -> ()
    | Dev d ->
        (Devices.flush d ~queue_id: 0 ();
         Devices.flush d ~queue_id: 1 ();
         to_cpu vect ();
         Devices.flush d ~queue_id: 0 ()));
   let s = ref seed
   in
     (for i = 0 to vect.Vector.length - 1 do
        s := f !s (Mem.unsafe_get vect i)
      done;
      !s))
  
let fold_right f vect seed =
  ((match vect.dev with
    | No_dev -> ()
    | Dev d ->
        (Devices.flush d ~queue_id: 0 ();
         Devices.flush d ~queue_id: 1 ();
         to_cpu vect ();
         Devices.flush d ~queue_id: 0 ()));
   let s = ref seed
   in
     (for i = vect.Vector.length - 1 downto 0 do
        s := f (Mem.unsafe_get vect i) !s
      done;
      !s))
  
let vfalse = Custom trueCustom
  
let vtrue = Custom falseCustom
  
let spoc_bool = vfalse
  
type vbool = bool Vector.vcustom

module Gpu =
  struct
    let gpuSum (var0 : bool Vector.vcustom) (var1 : Vector.vint32)
               (var2 : Vector.vint32) =
      Kernel.exec "kernels/Spoc_kernels" "sum"
        [| Kernel.VCustom (Kernel.relax_vcustom var0); Kernel.VInt32 var1;
          Kernel.VInt32 var2; Kernel.Int32 var0.Vector.length
        |]
      
    
		let sum (var0 : vbool) dev (block,grid)=
      let res = Vector.create Vector.int32 1 
			and tmp = Vector.create Vector.int32 (var0.Vector.length)
      in
				let tmp2 = grid.Kernel.gridX in 
					grid.Kernel.gridX <- (var0.Vector.length +  block.Kernel.blockX - 1) / block.Kernel.blockX ;
      	Kernel.run dev (block, grid) (gpuSum var0 res tmp);
				grid.Kernel.gridX <- tmp2;
				Mem.to_cpu res ();
				Devices.flush dev();
				Int32.to_int (Mem.get res 0)
      
		let int_bubble_filter (var0:bool Vector.vcustom) (var1: Vector.vint32) (var2: Vector.vint32)=
				      Kernel.exec "kernels/Spoc_kernels" "int_bubble_filter"
 [| Kernel.VCustom (Kernel.relax_vcustom var0); Kernel.VInt32 var1;
          Kernel.VInt32 var2;Kernel.Int32 var1.Vector.length
        |]
			

						
    let filter (dev:Devices.device) (block,grid) vect f =
      let vecBool = map f vfalse vect
      in
        ((match vecBool.dev with
          | Dev d when d = dev -> ()
          | _ -> Mem.to_device vecBool ~queue_id: 0 dev);
				Devices.flush dev();
         	let s2 = ref 0 in
						(iter 
							(fun elt -> 
									if (elt) then incr s2) vecBool;
							print !s2;
						let s = (*!s2 in*)
						sum vecBool dev (block,grid)
         		in
						print s;
						print (s - !s2);
           (
						let c = Vector.create vect.kind (s) in
            let idx = ref 0
            in
              (iteri
                 (fun elt i ->
                    if Mem.unsafe_get vecBool i
                    then (Mem.unsafe_set c !idx elt; incr idx)
                    else ())
                 vect;
               c))))
      
      
    let maskFilter dev ((block:Kernel.block),(grid:Kernel.grid)) vect (vecBool : bool Vector.vcustom)  =
      (if vect.length > vecBool.length
       then failwith "bool vector too small"
       else ()
			);
       (match vecBool.dev with
        | Dev d when d = dev -> ()
        | _ -> Mem.to_device vecBool ~queue_id: 0 dev
				);
       (match vect.dev with
        | Dev d when d = dev -> ()
        | _ -> Mem.to_device vect ~queue_id: 1 dev
			);
					let s2 = ref 0 in
						(iter 
							(fun elt -> 
									if (elt) then incr s2) vecBool;
							print !s2;
						let s = sum vecBool dev (block,grid)
         		in
						(print s;
						print (s - !s2);
          let c = Vector.create vect.kind vect.Vector.length in
						let tmp = grid.Kernel.gridX in 
							grid.Kernel.gridX <- 1;
					Kernel.run dev (block,grid) (int_bubble_filter vecBool vect c);
					grid.Kernel.gridX <- tmp;
					Mem.to_cpu c ();
					Devices.flush dev ();
					let res =
						Mem.sub_vector c 0 s in
					res))
					(* let idx = ref 0
          in
            (iteri
               (fun elt i ->
                  if Mem.get vecBool i
                  then (Mem.set c !idx elt; incr idx)
                  else ())
               vect;
             c)))*)
						
      
  end
  
