open Devices

type device_vec


type customarray

type 'a custom =
	{
		elt :'a;
		size : int ;
		get: customarray  ->  int -> 'a;
		set: customarray  ->  int -> 'a -> unit
	}




type  ('a, 'b) kind = 
	 		Float32 of ('a, 'b) Bigarray.kind
		| Float64 of ('a, 'b) Bigarray.kind
		| Int32 of ('a, 'b) Bigarray.kind
		| Int64 of ('a, 'b) Bigarray.kind
		| Complex32 of ('a, 'b) Bigarray.kind
		| Custom of 'a custom




(* shortcuts *)
let int32 = Int32 (Bigarray.int32)
let int64 = Int64 (Bigarray.int64)
let float32 = Float32 (Bigarray.float32)
let float64 = Float64 (Bigarray.float64)
let complex32 = Complex32 (Bigarray.complex32)


type ('a,'b)  spoc_vec = 
	| Bigarray of ('a, 'b, Bigarray.c_layout)Bigarray.Array1.t
	| CustomArray of (customarray * 'a custom)


external float32_of_float : float -> float = "float32_of_float"
external float_of_float32 : float -> float = "float_of_float32"


type vec_device =
	| No_dev
	| Dev of device

type ('a,'b) vector = {
	mutable device : int; (* -1 -> CPU, 0+ -> device *) 
	vector : ('a,'b) spoc_vec;(*('a, Bigarray.c_layout) Bigarray.Array1.t;(*('a,'b)spoc_vect;*)*)
	cuda_device_vec : device_vec array;
	opencl_device_vec : device_vec array;
	length : int;
	mutable dev:vec_device;
	kind: ('a, 'b)kind;
	}
	
external init_cuda_device_vec: unit -> device_vec  = "spoc_init_cuda_device_vec"
external init_opencl_device_vec : unit -> device_vec = "spoc_init_opencl_device_vec"


external create_custom : 'a custom  -> int -> customarray = "spoc_create_custom"


let create kind size =
			match kind with
			Float32  x  -> 			
			{ device = -1;
					vector = Bigarray (Bigarray.Array1.create x Bigarray.c_layout size) ; 
					cuda_device_vec = Array.create (!cuda_compatible_devices+1) (init_cuda_device_vec ());
					opencl_device_vec = Array.create (!opencl_compatible_devices+1) (init_opencl_device_vec ());
					length = size;
					dev = No_dev;
					kind = kind;
			}
			|	Float64  x->				
			{ device = -1;
					vector = Bigarray (Bigarray.Array1.create x Bigarray.c_layout size);
					cuda_device_vec = Array.create (!cuda_compatible_devices+1) (init_cuda_device_vec ());
					opencl_device_vec = Array.create (!opencl_compatible_devices+1) (init_opencl_device_vec ());
					length = size;
					dev = No_dev;
					kind = kind;
			}
			| Int32  x ->  
			{ device = -1;
							vector =  Bigarray (Bigarray.Array1.create x Bigarray.c_layout size);
							cuda_device_vec = Array.create (!cuda_compatible_devices+1) (init_cuda_device_vec ());
							opencl_device_vec = Array.create (!opencl_compatible_devices+1) (init_opencl_device_vec ());
							length = size;
							dev = No_dev;
							kind = kind;
						}
			| Complex32 x  ->   
			{ device = -1;
							vector =  Bigarray (Bigarray.Array1.create x Bigarray.c_layout size);
							cuda_device_vec = Array.create (!cuda_compatible_devices+1) (init_cuda_device_vec ());
							opencl_device_vec = Array.create (!opencl_compatible_devices+1) (init_opencl_device_vec ());
							length = size;
							dev = No_dev;
												kind = kind;

						}
			| Int64  x ->  
			{ device = -1;
							vector =  Bigarray (Bigarray.Array1.create x Bigarray.c_layout size);
							cuda_device_vec = Array.create (!cuda_compatible_devices+1) (init_cuda_device_vec ());
							opencl_device_vec = Array.create (!opencl_compatible_devices+1) (init_opencl_device_vec ());
							length = size;
							dev = No_dev;
							kind = kind;
						}
			| Custom c -> 
					{
					device = -1;
					vector =   CustomArray ((create_custom c size), c);
					cuda_device_vec = Array.create (!cuda_compatible_devices+1) (init_cuda_device_vec ());
					opencl_device_vec = Array.create (!opencl_compatible_devices+1) (init_opencl_device_vec ());
					length = size;
					dev = No_dev;
					kind = kind;

					}

	let length v = 
		v.length


type vfloat32 = (float, Bigarray.float32_elt)vector
type localvfloat32 = (float, Bigarray.float32_elt)vector
type vfloat64 = (float, Bigarray.float64_elt)vector
type vint32 = (int32, Bigarray.int32_elt)vector
type vint64 = (int64, Bigarray.int64_elt)vector
type vcomplex32 = (Complex.t, Bigarray.complex32_elt) vector
type  'a vcustom = ('a, 'a) vector


