open Camlp4.PreCast
open Syntax
open Ast
let expr_of_string = Syntax.Gram.parse_string Syntax.expr_eoi;;


let id = ref 0

let create_new_var loc =
	let v = "var"^(string_of_int !id) 
	in 
	let a = expr_of_string loc (v) in
	incr id;
	a,v


let rec string_of_ident = function
  | <:ident< $lid:s$ >> -> s
  | <:ident< $uid:s$ >> -> s
  | <:ident< $i1$.$i2$ >> -> "acc_" ^ (string_of_ident i1) ^ "_" ^ (string_of_ident i2)
  | <:ident< $i1$($i2$) >> -> "app_" ^ (string_of_ident i1) ^ "_" ^ (string_of_ident i2)
  | <:ident< $anti:_$ >> -> assert false

let fst = fun (x,y,z) -> x
let snd = fun (x,y,z) -> y
let thrd = fun (x,y,z) -> z

let rec parse_ctyp = function
	| TyNil _ -> failwith "TyNil";
	| TyAli _ -> failwith "TyAli";
	| TyAny _ -> failwith "TyAny";
	| TyApp (l, c1, c2) as t -> 
		let e,v =create_new_var l in
		let l1 = parse_ctyp c1 
		and l2 =parse_ctyp c2 in
		[ExTyc (l, e, t)],[v],[t] 
		(*failwith "TyApp";*)
	| TyArr (l,a,b) -> 
		(*Printf.printf "TyArr\n";*) 
		let l1 = parse_ctyp a 
		and l2 =parse_ctyp b in
			((fst l1)@(fst l2),(snd l1)@(snd l2),(thrd l1)@(thrd l2))
	| TyCls _ -> failwith "TyCls";
	| TyLab _ -> failwith "TyLab";
	| (TyId (l,id)) as t -> let e,v =create_new_var l in
				begin
					match t with
					| TyId(_, id) when (String.compare (string_of_ident id) "int32") = 0
						-> ([ExTyc (l,e,TyId(l, ident_of_expr (expr_of_string l "int")))],[v],[t])
					| TyId(_, id) when (String.compare (string_of_ident id) "int64") = 0
							-> ([ExTyc (l,e,TyId(l, ident_of_expr (expr_of_string l "int")))],[v],[t])
					| TyId(_, id) when (String.compare (string_of_ident id) "float32") = 0
						-> ([ExTyc (l,e,TyId(l, ident_of_expr (expr_of_string l "float")))],[v],[t])
					| TyId(_, id) when (String.compare (string_of_ident id) "float64") = 0
							-> ([ExTyc (l,e,TyId(l, ident_of_expr (expr_of_string l "float")))],[v],[t])		
					| _ ->  ([ExTyc (l,e,t)],[v],[t])
				end;
	| _ -> failwith "parse_ctyp"
	
	
let arg_string = ref ""
	
let rec apply_of_constaints loc file_name func_name id_list arg_list = function
	| [] ->  failwith "apply_of_constraints [] "
	(*| (ExApp (l, e1, e2))::q  -> 
		let f = PaTyc(l, PaId (loc, (ident_of_expr var)), typ) and
			f2 = apply_of_constaints loc file_name func_name id_list arg_list q in
				let _loc = loc in
					<:expr< fun $f$ -> $f2$ >>*)
	| (ExTyc (l,var,typ))::q -> 
			begin
				match q with 
				| t::(q2::q3) ->
					let f = PaTyc(l, PaId (loc, (ident_of_expr var)), typ) and
					f2 = apply_of_constaints loc file_name func_name id_list arg_list q in
					let _loc = loc in
					<:expr< fun $f$ -> $f2$ >>
				| _ -> let _loc = loc in
					let f = PaTyc(l, PaId (loc, (ident_of_expr var)), typ)in 
					let idx = ref 1 in
					let rec parseTyp t i= 
						match t with
							| TyId (_, id) when (String.compare (string_of_ident id) "int") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Int32 "^i^"; "^(!arg_string); 
							| TyId (_, id) when (String.compare (string_of_ident id) "int32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Int32 "^i^"; "^(!arg_string); 
							| TyId (_, id) when (String.compare (string_of_ident id) "int64") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Int64 "^i^"; "^(!arg_string);
							| TyId (_, id) when (String.compare (string_of_ident id) "float") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Float32 "^i^"; "^(!arg_string); 
							| TyId (_, id) when (String.compare (string_of_ident id) "float32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Float32 "^i^"; "^(!arg_string); 
							| TyId (_, id) when (String.compare (string_of_ident id) "float64") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Float64 "^i^"; "^(!arg_string); 
							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vfloat32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VFloat32 "^i^"; "^(!arg_string);
							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_localvfloat32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.LocalFloat32 "^i^"; "^(!arg_string);
							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vfloat64") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VFloat64 "^i^"; "^(!arg_string);  
							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vcomplex32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VComplex32 "^i^"; "^(!arg_string);
							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vint32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VInt32 "^i^"; "^(!arg_string); 
							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vint64") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VInt64 "^i^"; "^(!arg_string); 
							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vbool") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VCustom  (Spoc.Kernel.relax_vcustom "^i^"); "^(!arg_string);
							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vcustom") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VCustom  (Spoc.Kernel.relax_vcustom "^i^"); "^(!arg_string); 
							| TyId (_, id) -> 
								(*Printf.printf "--- %s\n" (string_of_ident id); 
								flush stdout;  *)
								arg_string := i^"; "^(!arg_string);
								(*failwith "apply_of_constraints TyId "*)
							| TyApp (l, t1, t2)  -> 
								begin
									match t2 with 
									| TyId (_, id) ->
										(*incr idx;*)
										(parseTyp t1 i );(*arg_string:= (string_of_ident id)^" "^ !arg_string;*)
									| _  -> failwith "erf"
								end
							| _ -> 
								failwith "apply_of_constraints Not TyId "
					in
					List.iter (fun i -> parseTyp  (List.nth id_list !idx ) i
							)
							 (List.tl (List.rev arg_list));
					let s = Printf.sprintf "Spoc.Kernel.exec \"%s\" \"%s\" [|%s|]\n" file_name func_name !arg_string in
					let exp = expr_of_string _loc s in
						arg_string := "";
					<:expr< fun $f$ -> $exp$ >>

			end
		| _ -> failwith "apply_of_constraints _ "
(*			ExApp (l, ExTyc(l, PAId (loc, (ident_of_expr var)), typ), apply_of_constaints loc q)*)

	



EXTEND Gram
  GLOBAL: str_item ;
  str_item: 
   [ [ "kernel"; lid = ident; ":";  typ = ctyp; "=";  file_name = STRING; func_name = STRING  -> 
					let constraint_var, arg_list, id_list  = parse_ctyp typ in
					let pat = apply_of_constaints _loc file_name func_name (List.rev id_list) arg_list constraint_var in
					let p1 = PaId(_loc, lid) in
					 let pPat =  
						<:binding< $p1$ = $pat$>>
					in
					(*ExApp (_loc, ExId(_loc, lid),  pat) in*)
							StVal	(_loc, ReNil, pPat)(*(_loc, (ExpTyc(, expr_of_string _loc "Spoc.Kernel.exec \n"))*)
									]];
	END